package phyApps;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Label;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Timer;
import java.util.TimerTask;

import javax.swing.JOptionPane;

import phyApps.common.Problem;
import phyApps.common.ControlPanel;
import phyApps.common.ControlPanelListener;
import phyApps.common.Result;
import phyApps.common.graph.ArrowPlot;
import phyApps.common.graph.AxisPlot;
import phyApps.common.graph.CoordinateSystem;
import phyApps.common.graph.Graph;
import phyApps.common.graph.Plot;
import phyApps.common.graph.Point;
import phyApps.common.graph.Region;
import phyApps.common.mvc.Controller;
import phyApps.common.mvc.Model;
import phyApps.common.mvc.View;
import phyApps.common.mvc.PhyApplet;

@SuppressWarnings("serial")
public class Cart extends PhyApplet {
	
	public static double LENGTH = 2;
	
	public void init() {
		setupMVC(new CartModel(), new CartView(), new CartController());
	}
	
	public static void main(String[] args) {
		PhyApplet.runApplet(new Cart(), "Inclined Plane");
	}
}

class CartModel extends Model {
	public static double LENGTH = Cart.LENGTH;
	private static int PERIOD = 25;
	private static double NUDGE_VELOCITY = 0.1;

	private double gravity;
	private double mass;
	private double angle; // In degrees
	private double staticFriction; // coefficient of static friction
	private double kineticFriction; // coefficient of kinetic friction
	private double frictionForce;
	private double normalForce;
	private double gravityForce;
	private double force;
	private double position;
	private double velocity;
	private double acceleration;
	private double totalTime; // The time it will take for the cart to reach the bottom of the ramp.
	
	private Timer timer;
	private int time; // The amount of time passed so far, in milliseconds.
	private boolean moving;
	
	public CartModel() {
		moving = false;
	}
	
	private void stop() {
		if (timer != null) {
			timer.cancel();
			timer = null;
			velocity = 0.0;
		}
	}
	
	public void start() {
		// Timer must be null.
		timer = new Timer();
		timer.scheduleAtFixedRate(new CartTask(), 0, PERIOD);
		moving = true;
	}
	
	public void setup(double mass, double angle, double staticFriction, double kineticFriction, double gravity, boolean redraw) {
		// Stop the simulation, if it is running.
		stop();
		moving = false;
		
		this.gravity = gravity;
		this.mass = mass;
		this.angle = angle;
		this.staticFriction = staticFriction;
		this.kineticFriction = kineticFriction;

		time = 0;
		position = 0;
		velocity = 0;
		calculateForces();
		
		if (redraw)
			draw();
	}

	public void nudge() {
		if (timer != null) {
			velocity = velocity + NUDGE_VELOCITY;
			calculateForces();
			moving = true;
		}
	}
	
	public void tick(int time) {
		if (velocity <= 0.0)
			calculateForces();
		velocity += acceleration * PERIOD / 1000.0;
		if (velocity < 0.0)
			velocity = 0.0;
		position += velocity * PERIOD / 1000.0;
		if (position > LENGTH) {
			stop();
			position = LENGTH;
		}
		draw();
	}
	
	class CartTask extends TimerTask {
		public void run() {
			time = time + PERIOD;
			tick(time);
		}
	}
	
	/* For the View/Controller: */
	
	public double getAngle(){
		return angle;
	}
	
	public double getFrictionForce(){
		return frictionForce;
	}
	
	public double getNormalForce(){
		return normalForce;
	}
	
	public double getGravityForce(){
		return gravityForce;
	}
	
	public double getForce(){
		return force;
	}
	
	public double getMass() {
		return mass;
	}
	
	public double getPosition(){
		return position;
	}
	
	public double getAcceleration(){
		return acceleration;
	}
	
	public double getTotalTime(){
		return totalTime;
	}
	
	public boolean isMoving() {
		return moving;
	}
	
	/* Calculations */
	
	public void calculateForces() {
		gravityForce = mass * gravity;
		normalForce = gravityForce * Math.cos((Math.PI/180.0)*angle);
		double staticFrictionForce = staticFriction * normalForce;
		double parallelForce = gravityForce * Math.sin((Math.PI/180.0)*angle);
		if (velocity > 0.0 || parallelForce > staticFrictionForce) {
			frictionForce = kineticFriction * normalForce;
			force = parallelForce - frictionForce;
		}
		else {
			frictionForce = parallelForce;
			force = 0.0;
		}
		acceleration = force/mass;
		totalTime = Math.sqrt(2 * LENGTH / acceleration);
	}
	
	public static double calculateAcceleration(double gravity, double angle) {
		return gravity * Math.sin((Math.PI/180.0)*angle);
	}
	
	public static double calculateKineticFriction(double gravity, double angle, double acceleration) {
		return Math.tan((Math.PI/180.0)*angle) - acceleration / gravity / Math.cos((Math.PI/180.0)*angle);
	}
	
	public static double calculateKineticFriction(double gravity, double angle, double length, double time) {
		return calculateKineticFriction(gravity, angle, 2.0 * length / time / time);
	}
	
	public static double calculateStaticFriction(double angle) {
		return Math.tan((Math.PI/180.0)*angle);
	}
	
	public static double calculateFinalVelocity(double length, double acceleration) {
		return Math.sqrt(2.0 * acceleration * length);
	}
	
	public static double calculateTime(double gravity, double angle, double length, double kineticFriction) {
		return Math.sqrt(2.0 * length / (gravity * (Math.sin((Math.PI/180.0)*angle) - kineticFriction * Math.cos((Math.PI/180.0)*angle))));
	}
}


@SuppressWarnings("serial")
class CartController extends Controller implements ActionListener, ControlPanelListener {

	private CartModel model;
	private ControlPanel controlPanel;
	
	private String HINT_TITLE = "Equations";
	private String HINT_TEXT = "<html><br>" +
		"&nbsp;&nbsp;Here are some equations you may find helpful:&nbsp;&nbsp;<br><br>" +
		"&nbsp;&nbsp;Force of gravity<br>" +
		"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Fg = m * g<br><br>" +
		"&nbsp;&nbsp;Normal force<br>" +
		"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Fn = Fg * cos(\u0398)<br><br>" +
		"&nbsp;&nbsp;Force of friction<br>" +
		"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Ff = \u03bc * Fn<br><br>" +
		"&nbsp;&nbsp;Total force<br>" +
		"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Ftot = Fg * sin(\u0398) - Ff" +
		"<br></html>";
	private String HELP_TITLE = "Help";
	public static String HELP_TEXT =
		"Things to know:\n\n" +
		"- After changing a text field, press enter to apply the change.\n\n" +
		"- The ramp is 2.5 meters long and the cart is 0.5 meters long,\n" +
		"so the total distance the cart travels is 2 meters.\n\n" +
		"- The vectors drawn on the cart are acceleration vectors, not force\n" +
		"vectors.  Remember, though, that force vectors are just acceleration\n" +
		"vectors scaled by the mass.\n\n" +
		"- In real systems, the coefficient of static friction is generally\n" +
		"higher than the coefficient of kinetic friction, so the simulation\n" +
		"enforces this constraint.\n\n" +
		"- If the static friction is too high, the cart won't move when\n" +
		"you release it.  But there is a 'nudge' button to give it a\n" +
		"small initial velocity.";
	
	public Component getComponent() {
		return this;
	}

	@Override
	public void setup(View view, Model model) {
		this.model = (CartModel) model;
		
		controlPanel = new ControlPanel();
		
		controlPanel.addScrollbar("angle", "Angle (deg)", 30, 0, 90, 1);
		controlPanel.addScrollbar("static", "Static Friction", 0.5, 0.0, 2.0, 0.01);
		controlPanel.addScrollbar("kinetic", "Kinetic Friction", 0.4, 0.0, 2.0, 0.01);
		controlPanel.addTextField("mass", "Mass (g)", 200);
		controlPanel.addTextField("g", "Gravity (m/s/s)", 9.81);
		
		controlPanel.addLabel("gravity", "Force of Gravity", "N");
		controlPanel.setColor("gravity", CartView.GRAVITY_COLOR);
		controlPanel.addLabel("friction", "Force of Friction", "N");
		controlPanel.setColor("friction", CartView.FRICTION_COLOR);
		controlPanel.addLabel("normal", "Normal Force", "N");
		controlPanel.setColor("normal", CartView.NORMAL_COLOR);
		controlPanel.addLabel("force", "Total Force", "N");
		controlPanel.setColor("force", CartView.FORCE_COLOR);
		controlPanel.addLabel("acceleration", "Acceleration", "m/s/s");
		controlPanel.addLabel("time", "Total Time", "s");
		
		controlPanel.addButton("setup", "Setup", this);
		controlPanel.addButton("release", "Release", this);
		controlPanel.addButton("nudge", "Nudge", this);
		
		controlPanel.addText("");
		controlPanel.addButton("hint", HINT_TITLE, this);
		controlPanel.addButton("help", HELP_TITLE, this);
		
		controlPanel.addText("");
		controlPanel.addText("Problems", Label.CENTER);

		controlPanel.addButton("problem1", "Frictionless Acceleration", this);
		controlPanel.addButton("problem2", "Acceleration", this);
		controlPanel.addButton("problem3", "Kinetic Friction", this);
		controlPanel.addButton("problem4", "Static Friction", this);
		controlPanel.addButton("problem5", "Which Planet?", this);
		
		controlPanel.addListener("angle", this);
		controlPanel.addListener("mass", this);
		controlPanel.addListener("static", this);
		controlPanel.addListener("kinetic", this);
		controlPanel.addListener("g", this);
		
		add(controlPanel);
		
		setup();
	}
	
	public void setup() {
		double mass = controlPanel.getValue("mass");
		double angle = controlPanel.getValue("angle");
		double staticFriction = controlPanel.getValue("static");
		double kineticFriction = controlPanel.getValue("kinetic");
		double g = controlPanel.getValue("g");
		
		if (g < 0) {
			controlPanel.setValue("g", 0.0);
			g = 0.0;
		}
		if (mass < 0) {
			controlPanel.setValue("mass", 0.0);
			mass = 0.0;
		}
		
		model.setup(mass, angle, staticFriction, kineticFriction, g, true);
	}
	
	// angle in degrees
	public void setup(double mass, double angle, double staticFriction, double kineticFriction, double grav) {
		controlPanel.setValue("mass", mass);
		controlPanel.setValue("angle", angle);
		controlPanel.setValue("static", staticFriction);
		controlPanel.setValue("kinetic", kineticFriction);
		controlPanel.setValue("g", grav);
		setup();
	}

	public void update() {
		controlPanel.setValue("friction", model.getFrictionForce());
		controlPanel.setValue("normal", model.getNormalForce());
		controlPanel.setValue("gravity", model.getGravityForce());
		controlPanel.setValue("force", model.getForce());
		controlPanel.setValue("acceleration", model.getAcceleration());
		controlPanel.setValue("time", model.getTotalTime());
	}
	
	public void actionPerformed(ActionEvent event) {
		if (event.getActionCommand() == "setup")
			setupButton();
		else if (event.getActionCommand() == "nudge")
			nudgeButton();
		else if (event.getActionCommand() == "release")
			releaseButton();
		else if (event.getActionCommand().startsWith("problem"))
			problemButton(event.getActionCommand());
		else if (event.getActionCommand() == "hint")
			Problem.popupFrame(HINT_TITLE, HINT_TEXT);
		else if (event.getActionCommand() == "help")
			JOptionPane.showMessageDialog(this, HELP_TEXT, HELP_TITLE, JOptionPane.PLAIN_MESSAGE);
	}

	private void setupButton() {
		setup();
	}
	
	private void releaseButton() {
		setup();
		model.start();
	}
	
	private void nudgeButton() {
		if (!model.isMoving()) {
			setup();
			model.start();
			model.nudge();
		}
		else {
			model.nudge();
		}
	}
	
	private void problemButton(String problemName) {
		Problem problem = generateProblem(problemName);
		Frame frame = Problem.popupFrame(HINT_TITLE, HINT_TEXT);
		problem.show(null);
		frame.dispose();
		problem.setup();
	}
	
	public Problem generateProblem(String problemNumber) {
		Problem problem = null;
		if (problemNumber == "problem1")
			problem = new FrictionlessAccelerationProblem(this);
		else if (problemNumber == "problem2")
			problem = new AccelerationProblem(this);
		else if (problemNumber == "problem3")
			problem = new KineticFrictionProblem(this);
		else if (problemNumber == "problem4")
			problem = new StaticFrictionProblem(this);
		else if (problemNumber == "problem5")
			problem = new PlanetProblem(this);
		return problem;
	}
	
	public void setGravity(double g) {
		controlPanel.setValue("g", g);
	}

	public void setMass(double mass) {
		controlPanel.setValue("mass", mass);
	}
	
	public void setAngle(double angle) {
		controlPanel.setValue("angle", angle);
	}
	
	public void setStaticFriction(double friction) {
		controlPanel.setValue("static", friction);
	}
	
	public void setKineticFriction(double friction) {
		controlPanel.setValue("kinetic", friction);
	}

	public void valueChanged(String field, double value) {
		if (field == "kinetic") {
			if (controlPanel.getValue("static") < value)
				controlPanel.setValue("static", value);
		}
		else if (field == "static") {
			if (controlPanel.getValue("kinetic") > value)
				controlPanel.setValue("kinetic", value);
		}
		this.setupButton();
	}
}


@SuppressWarnings("serial")
class CartView extends View {
	
	private static double SCALE = 250.0;
	private static double VECTOR_SCALE = 0.05;
	public static double CART_LENGTH = 0.50;
	public  static double CART_HEIGHT = 0.3;
	public static int RAMP_THICKNESS = 10;
	private static int TOP_BORDER = 0;
	private static double RIGHT_BORDER = 0.5;
	public static int RAMP_OFFSET = 0;
	
	public static double LENGTH = CartModel.LENGTH + CART_LENGTH;
	
	public static Color GRAVITY_COLOR = new Color(0, 0, 220);
	public static Color NORMAL_COLOR = new Color(150, 0, 100);
	public static Color FRICTION_COLOR = new Color(200, 0, 0);
	public static Color FORCE_COLOR = new Color(0, 175, 0);
	
	private Graph graph;
	private CartPlot cartPlot;
	private ArrowPlot gravityArrow;
	private ArrowPlot normalArrow;
	private ArrowPlot frictionArrow;
	private ArrowPlot forceArrow;

	public CartView() {
		graph = new Graph(new Dimension((int) (SCALE*(LENGTH + RIGHT_BORDER)), (int) (SCALE*(LENGTH + TOP_BORDER + RAMP_OFFSET))));
		cartPlot = new CartPlot(graph);
		AxisPlot axisPlot = new AxisPlot(graph);
		axisPlot.toggleScale(true);
		axisPlot.toggleNumbers(true);
		axisPlot.setXUnit("m");
		axisPlot.setYUnit("m");
		cartPlot.setRegion(new Region(0, 0, LENGTH + RIGHT_BORDER, LENGTH + TOP_BORDER + RAMP_OFFSET));
		gravityArrow = new ArrowPlot(graph, GRAVITY_COLOR, null);
		normalArrow = new ArrowPlot(graph, NORMAL_COLOR, null);
		frictionArrow = new ArrowPlot(graph, FRICTION_COLOR, null);
		forceArrow = new ArrowPlot(graph, FORCE_COLOR, null);
	}
	
	public void draw(Model m) {
		CartModel model = (CartModel) m;
		
		double position = model.getPosition();
		double friction = model.getFrictionForce();
		double gravity = model.getGravityForce();
		double normal = model.getNormalForce();
		double force = model.getForce();
		double angle = model.getAngle();
		double mass = model.getMass();
		
		cartPlot.setAngle(angle);
		cartPlot.setPosition(position);
		
		// Draw the force vectors.
		
		Point start = new Point(LENGTH, 0);
		Point end = Point.multiply(LENGTH, new Point(1.0 - Math.cos((Math.PI/180.0)*angle), Math.sin((Math.PI/180.0)*angle)));
		Point half_height = Point.multiply(CART_HEIGHT/2,
				new Point(Math.sin((Math.PI/180.0)*angle), Math.cos((Math.PI/180.0)*angle)));
		Point bottom = Point.sum(end, Point.multiply((position + CART_LENGTH/2.0)/LENGTH,
				Point.sum(start, Point.multiply(-1.0, end))));
		Point center = Point.sum(bottom, half_height);
		
		Point gravityVector = Point.polarToCartesian(VECTOR_SCALE * gravity / mass, -Math.PI/2.0);
		Point normalVector = Point.polarToCartesian(VECTOR_SCALE * normal / mass, Math.PI/2.0 - Math.PI*angle/180.0);
		Point frictionVector = Point.polarToCartesian(VECTOR_SCALE * friction / mass, Math.PI - Math.PI*angle/180.0);
		Point forceVector = Point.polarToCartesian(VECTOR_SCALE * force / mass, -Math.PI*angle/180.0);
		
		if (!model.isMoving()) {
			gravityArrow.move(center, Point.sum(center, gravityVector));
			//normalArrow.move(Point.sum(center, Point.multiply(-1.0, normalVector)), center);
			normalArrow.move(center, Point.sum(center, normalVector));
			frictionArrow.move(center, Point.sum(center, frictionVector));
			forceArrow.move(center, Point.sum(center, forceVector));
		}
		else {
			gravityArrow.clear();
			normalArrow.clear();
			frictionArrow.clear();
			//forceArrow.clear();
			forceArrow.move(center, Point.sum(center, forceVector));
		}
		
		graph.draw();
	}

	@Override
	public Component getComponent() {
		return graph;
	}
	
}

class CartPlot extends Plot {

	private static double LENGTH = CartView.LENGTH;
	private double position;
	private double angle;
	
	private static Color RAMP_COLOR = new Color(145, 90, 45);
	private static Color CART_COLOR = Color.BLACK;
	
	public CartPlot(Graph graph) {
		super(graph);
	}

	public Region determineRegion() {
		return null;
	}

	public void draw(Graphics g) {
		CoordinateSystem space = graph.getCoordinateSpace();
		
		// Draw the ramp.
		Point start = new Point(LENGTH, CartView.RAMP_OFFSET);
		Point end = Point.multiply(LENGTH, new Point(1.0 - Math.cos((Math.PI/180.0)*angle), CartView.RAMP_OFFSET / LENGTH + Math.sin((Math.PI/180.0)*angle)));
		graph.drawLine(g, start, end, RAMP_COLOR, CartView.RAMP_THICKNESS);
		
		// Draw the cart.
		// a helpful vector:
		Point height = Point.polarToCartesian(CartView.CART_HEIGHT, Math.PI/2.0 - Math.PI*angle/180.0);
		
		// Draw a stopper at the end of the ramp.
		Point S = Point.sum(start, height);
		Point T = start;
		int y = (int) (CartView.RAMP_THICKNESS * Math.sin((Math.PI/180.0)*angle) / 2.0);
		int x = (int) (CartView.RAMP_THICKNESS * Math.cos((Math.PI/180.0)*angle) / 2.0);
		S = space.translate(S, x, y);
		T = space.translate(T, x, y);
		graph.drawLine(g, S, T, RAMP_COLOR, CartView.RAMP_THICKNESS);
		
		// The four corners of the cart.
		Point A, B, C, D;
		
		double Ascale = (LENGTH - position)/LENGTH;
		double Bscale = (LENGTH - position - CartView.CART_LENGTH)/LENGTH;
		A = new Point(start.getX() + Ascale * end.getX() - Ascale * start.getX(),
				start.getY() + Ascale * end.getY() - Ascale * start.getY());
		B = new Point(start.getX() + Bscale * end.getX() - Bscale * start.getX(),
				start.getY() + Bscale * end.getY() - Bscale * start.getY());
		A = space.translate(A, y, -x);
		B = space.translate(B, y, -x);
		C = Point.sum(A, height);
		D = Point.sum(B, height);
		Point[] corners = {A, B, D, C};
		graph.fillPolygon(g, corners, CART_COLOR);
	}

	public void setAngle(double angle) {
		this.angle = angle;
	}
	
	public void setPosition (double position) {
		this.position = position;
	}
	
	public boolean isEmpty() {
		return false;
	}
}

abstract class CartProblem extends Problem {
	protected static String TITLE = "Inclined Ramp Problem";
	protected static double LENGTH = Cart.LENGTH;
	
	protected CartController controller;
	
	protected int mass;
	protected int angle;
	protected double kineticFriction;
	protected double staticFriction;
	protected double gravity;
	protected double acceleration;
}


class FrictionlessAccelerationProblem extends CartProblem {
	protected static String TEXT =
		"A cart of mass %d grams is placed on the top of a\n" +
		"frictionless ramp with an incline of %d degrees. \n" +
		"Find the acceleration of the cart down the ramp.";
	
	public FrictionlessAccelerationProblem(CartController controller) {
		this.controller = controller;
		mass = randInt(200, 300, 10);
		angle = randInt(15, 40, 1);
		gravity = 9.81;
		staticFriction = 0.0;
		kineticFriction = 0.0;
		acceleration = CartModel.calculateAcceleration(gravity, angle);
	}
	
	public Result display(Component parent) {
		String problemText = String.format(TEXT, mass, angle);
		double answer = popup(parent, problemText, TITLE);
		return checkAnswer(answer, acceleration);
	}

	public void setup() {
		controller.setup(mass, angle, staticFriction, kineticFriction, gravity);
	}
}


class AccelerationProblem extends CartProblem {
	protected static String TEXT =
		"A cart of mass %d grams is placed on the top of a\n" +
		"ramp of incline %d degrees. The coefficient of \n" +
		"kinetic friction between the cart and the ramp is %.2f.\n" +
		"Find the acceleration of the cart down the ramp.";
	
	public AccelerationProblem(CartController controller) {
		this.controller = controller;
		mass = randInt(250, 350, 10);
		angle = randInt(30, 45, 1);
		gravity = 9.81;
		acceleration = randInt(1, 4, 1);
		kineticFriction = CartModel.calculateKineticFriction(gravity, angle, acceleration);
		staticFriction = 1.05 * kineticFriction;
	}
	
	public Result display(Component parent) {
		String problemText = String.format(TEXT, mass, angle, kineticFriction);
		double answer = popup(parent, problemText, TITLE);
		return checkAnswer(answer, acceleration);
	}

	public void setup() {
		controller.setup(mass, angle, staticFriction, kineticFriction, gravity);
	}
}


class KineticFrictionProblem extends CartProblem {
	protected static String TEXT =
		"A cart of mass %d grams is placed on the top of a \n" +
		"%d meter long ramp inclined at %d degrees. The cart is \n" +
		"released and takes %d seconds to reach the end of the \n" +
		"ramp. Calculate the coefficient of kinetic friction \n" +
		"between the cart and the ramp.";
	
	protected int totalTime;
	
	public KineticFrictionProblem(CartController controller) {
		this.controller = controller;
		mass = randInt(250, 350, 10);
		angle = randInt(20, 45, 1);
		gravity = 9.81;
		acceleration = randInt(1, 5, 1);
		totalTime = randInt(2, 5, 1);
		kineticFriction = CartModel.calculateKineticFriction(gravity, angle, LENGTH, totalTime);
		staticFriction = kineticFriction;
	}
	
	public Result display(Component parent) {
		String problemText = String.format(TEXT, mass, (int) LENGTH, angle, totalTime);
		double answer = popup(parent, problemText, TITLE);
		return checkAnswer(answer, kineticFriction);
	}

	public void setup() {
		controller.setup(mass, angle, staticFriction, kineticFriction, gravity);
	}
}


class StaticFrictionProblem extends CartProblem {
	protected static String TEXT =
		"A cart of mass %d grams is placed on a horizontal ramp.\n" +
		"the incline of the ramp is slowly increased until the\n" +
		"cart starts to slide. The occurs at an incline of %d\n" +
		"degrees. Calculate the coefficient of static friction\n" +
		"between the cart and ramp.";
	
	public StaticFrictionProblem(CartController controller) {
		this.controller = controller;
		mass = randInt(250, 350, 10);
		angle = randInt(20, 60, 1);
		gravity = 9.81;
		staticFriction = CartModel.calculateStaticFriction(angle);
		kineticFriction = 0.8 * staticFriction;
	}
	
	public Result display(Component parent) {
		String problemText = String.format(TEXT, mass, angle);
		double answer = popup(parent, problemText, TITLE);
		return checkAnswer(answer, staticFriction);
	}

	public void setup() {
		controller.setup(mass, angle, staticFriction, 0.75 * staticFriction, gravity);
	}
}


class PlanetProblem extends CartProblem {
	protected static String TEXT =
		"You are on an outpost on some planet in the solar system,\n" +
		"but you don't know which. There are no windows. However,\n" +
		"you do have a cart and a ramp. The ramp is %d meters long\n" +
		"and the length of the cart is negligible in comparison.\n" +
		"When the ramp is angled at 30 degrees, the cart takes\n" +
		"%.2f seconds to reach the bottom. At 45 degrees, it takes\n" +
		"%.2f seconds. Which planet are you on?\n" +
		"Don't forget to take friction into account!";
	
	protected double time1;
	protected double time2;
	protected String[] planets = {"Select a planet", "Venus (g = 8.87 m/s/s)",
			"Earth (g = 9.78 m/s/s)", "Mars (g = 3.71 m/s/s)", "Uranus (g = 8.69 m/s/s)", "Neptune (g = 11.15 m/s/s)"};
	protected double[] planetGravities = {0.0, 8.87, 9.78, 3.71, 8.69, 11.15};
	protected int planet;
	
	public PlanetProblem(CartController controller) {
		this.controller = controller;
		planet = randInt(1, planets.length - 1, 1);
		kineticFriction = randDouble(0.2, 0.5, 0.1);
		gravity = planetGravities[planet];
		time1 = CartModel.calculateTime(gravity, 30, LENGTH, kineticFriction);
		time2 = CartModel.calculateTime(gravity, 45, LENGTH, kineticFriction);
	}
	
	public Result display(Component parent) {
		String problemText = String.format(TEXT, (int) LENGTH, time1, time2);
		String answer = popup(parent, problemText, TITLE, planets);
		if (answer == null)
			return Result.CANCEL;
		if (answer.equals(planets[planet]))
			return Result.CORRECT;
		else
			return Result.INCORRECT;
	}

	public void setup() {
		controller.setup(200, 30, 1.10 * kineticFriction, kineticFriction, gravity);
	}
}
