package phyApps;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Label;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Timer;
import java.util.TimerTask;

import javax.swing.BoxLayout;
import javax.swing.JOptionPane;

import phyApps.common.ControlPanel;
import phyApps.common.ControlPanelListener;
import phyApps.common.Problem;
import phyApps.common.Result;
import phyApps.common.graph.AxisPlot;
import phyApps.common.graph.CoordinateSystem;
import phyApps.common.graph.Graph;
import phyApps.common.graph.Point;
import phyApps.common.graph.PointPlot;
import phyApps.common.graph.Region;
import phyApps.common.mvc.Controller;
import phyApps.common.mvc.Model;
import phyApps.common.mvc.View;
import phyApps.common.mvc.PhyApplet;

@SuppressWarnings("serial")
public class CollisionApplet extends PhyApplet {
	
	public static double DISTANCE = 10.0;
	
	public void init() {
		setupMVC(new CollisionModel(), new CollisionView(), new CollisionController());
	}
	
	protected String getTitle() {
		return "Collisions";
	}
	
	public static void main(String[] args) {
		PhyApplet.runApplet(new CollisionApplet(), "Collision");
	}
}


@SuppressWarnings("serial")
class CollisionController extends Controller implements ActionListener, ControlPanelListener {
	
	CollisionModel model;
	CollisionView view;
	ControlPanel controlPanel;
	
	private static double INIT_MASS1 = 7;
	private static double INIT_MASS2 = 10;
	private static double INIT_VELOCITY1 = 1;
	private static double INIT_VELOCITY2 = -1;
	private static double INIT_RESTITUTION = 1.0;
	
	protected static double MIN_MASS = 1.0;
	protected static Font BOLD_FONT = new Font("sanserif", Font.BOLD, 12);
	protected static String MOMENTUM_UNITS = "N*s";
	
	private static String EQUATIONS_TITLE = "Equations";
	private static String HELP_TITLE = "Help";
	private static String EQUATIONS_TEXT = "<html><br>" +
		"&nbsp;&nbsp;Suppose two balls with masses M\u2081 and M\u2082 and velocities V\u2081 and V\u2082 collide&nbsp;&nbsp;<br>" +
		"&nbsp;&nbsp;in space, using the convention that positive velocity denotes motion to the<br>" +
		"&nbsp;&nbsp;right and negative velocity denotes motion to the left.  After the collision<br>" +
		"&nbsp;&nbsp;the balls will have new velocities Vf\u2081 and Vf\u2082.<br>" +
		"&nbsp;&nbsp;In any collision of this nature, the total momentum of the system is<br>" +
		"&nbsp;&nbsp;conserved.<br><br>" +
		"&nbsp;&nbsp;Before the collision, the speed at which the balls move toward each other<br>" +
		"&nbsp;&nbsp;is V\u2081 - V\u2082.  Call this the initial speed.  After the collision, the speed at which<br>" +
		"&nbsp;&nbsp;they move away from one another is Vf\u2082 - Vf\u2081.  Call this the final speed.  If<br>" +
		"&nbsp;&nbsp;both objects are completely elastic i.e. bouncy, then the final speed will<br>" +
		"&nbsp;&nbsp;be equal to the initial speed.  However, in an inelastic collision the final<br>" +
		"&nbsp;&nbsp;speed will be less than the initial speed because some energy is lost to the<br>" +
		"&nbsp;&nbsp;deformation of the balls.  The coefficient of restitution R between the two<br>" +
		"&nbsp;&nbsp;balls is defined as the ratio of the final speed to the initial speed.<br><br>" +
		"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;R = (V\u2081 - V\u2082) / (Vf\u2082 - Vf\u2081)<br><br>" +
		"&nbsp;&nbsp;Using the definition of r and the fact that momentum is conserved, you can<br>" +
		"&nbsp;&nbsp;derive the equation for the final velocity of the first ball.<br><br>" +
		"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Vf\u2081 = ((1 + R)*M\u2082*V\u2082 + V\u2081*(M\u2081 - R*M\u2082)) / (M\u2081 + M\u2082)<br><br>" +
		"&nbsp;&nbsp;If you are clever, the same expression can also be used to find the final<br>" +
		"&nbsp;&nbsp;velocity of the second ball.<br>" +
		"</html>";
	
	static String HELP_TEXT =
		"Things to know:\n\n" +
		"- After changing a text field, press enter to apply the change.\n\n" +
		"- The coefficient of restitution of an object is a measure of\n" +
		"its bounciness; specifically, it is the fraction of its velocity\n" +
		"that is conserved during a collision.";
	
	public CollisionController() {		
		controlPanel = new ControlPanel();
		
		Color COLOR_A = CollisionView.COLOR_A;
		Color COLOR_B = CollisionView.COLOR_B;
		
		controlPanel.addLabel("distance", "Distance", "m");
		controlPanel.setValue("distance", CollisionApplet.DISTANCE);
		controlPanel.addTextField("mass1", "Mass of A", INIT_MASS1, "kg", MIN_MASS, Double.NaN);
		controlPanel.setColor("mass1", COLOR_A);
		controlPanel.addTextField("mass2", "Mass of B", INIT_MASS2, "kg", MIN_MASS, Double.NaN);
		controlPanel.setColor("mass2", COLOR_B);
		controlPanel.addTextField("restitution", "Restitution", INIT_RESTITUTION, null, 0.0, 1.0);
		
		controlPanel.addText("");
		controlPanel.addText("Before Collision", Label.CENTER, BOLD_FONT);
		controlPanel.addTextField("velocity1", "Velocity A", INIT_VELOCITY1, "m/s");
		controlPanel.setColor("velocity1", COLOR_A);
		controlPanel.addTextField("velocity2", "Velocity B", INIT_VELOCITY2, "m/s");
		controlPanel.setColor("velocity2", COLOR_B);
		controlPanel.addTextField("momentum1", "Momentum A", INIT_MASS1 * INIT_VELOCITY1);
		controlPanel.setColor("momentum1", COLOR_A);
		controlPanel.addTextField("momentum2", "Momentum B", INIT_MASS2 * INIT_VELOCITY2);
		controlPanel.setColor("momentum2", COLOR_B);
		controlPanel.addLabel("momentum", "Total Momentum", MOMENTUM_UNITS);
		controlPanel.addLabel("energy", "Total Energy", "J");
		
		controlPanel.addText("");
		controlPanel.addText("After Collision", Label.CENTER, BOLD_FONT);
		controlPanel.addLabel("velocity1f", "Velocity A", "m/s");
		controlPanel.setColor("velocity1f", COLOR_A);
		controlPanel.addLabel("velocity2f", "Velocity B", "m/s");
		controlPanel.setColor("velocity2f", COLOR_B);
		controlPanel.addLabel("momentum1f", "Momentum A", MOMENTUM_UNITS);
		controlPanel.setColor("momentum1f", COLOR_A);
		controlPanel.addLabel("momentum2f", "Momentum B", MOMENTUM_UNITS);
		controlPanel.setColor("momentum2f", COLOR_B);
		controlPanel.addLabel("momentumf", "Total Momentum", MOMENTUM_UNITS);
		controlPanel.addLabel("energyf", "Total Energy", "J");
		
		controlPanel.addButton("start", "Start", this);
		controlPanel.addButton("equations", "Equations", this);
		controlPanel.addButton("help", "Help", this);
		
		controlPanel.addText("");
		controlPanel.addText("Problems", Label.CENTER, BOLD_FONT);
		controlPanel.addButton("problem1", "Elastic Collision", this);
		controlPanel.addButton("problem2", "Inelastic Collision", this);
		
		controlPanel.addListener("velocity1", this);
		controlPanel.addListener("velocity2", this);
		controlPanel.addListener("mass1", this);
		controlPanel.addListener("mass2", this);
		controlPanel.addListener("momentum1", this);
		controlPanel.addListener("momentum2", this);
		controlPanel.addListener("restitution", this);
		
		add(controlPanel);
	}
	
	public void setup(View view, Model model) {
		this.view = (CollisionView) view;
		this.model = (CollisionModel) model;
		this.model.setup(INIT_MASS1, INIT_VELOCITY1, INIT_MASS2, INIT_VELOCITY2, INIT_RESTITUTION);
		this.view.setup(this.model);
	}

	public void update() { // called by model.draw()
		Ball ball1 = model.getBall1();
		Ball ball2 = model.getBall2();
		double vel1 = model.getFinalVelocity(ball1);
		double vel2 = model.getFinalVelocity(ball2);
		
		controlPanel.setValue("momentum", ball1.mass * ball1.velocity + ball2.mass * ball2.velocity);
		controlPanel.setValue("energy", ball1.mass * ball1.velocity * ball1.velocity / 2.0 +
				ball2.mass * ball2.velocity * ball2.velocity / 2.0);
		controlPanel.setValue("energyf", ball1.mass * vel1 * vel1 / 2.0 + ball2.mass * vel2 * vel2 / 2.0);
		controlPanel.setValue("velocity1f", vel1);
		controlPanel.setValue("velocity2f", vel2);
		controlPanel.setValue("momentum1f", ball1.mass * vel1);
		controlPanel.setValue("momentum2f", ball2.mass * vel2);
		controlPanel.setValue("momentumf", ball1.mass * vel1 + ball2.mass * vel2);
	}
	
	public void setup(double mass1, double mass2, double vel1, double vel2, double restitution) {
		controlPanel.setValue("restitution", restitution);
		controlPanel.setValue("mass1", mass1);
		controlPanel.setValue("mass2", mass2);
		controlPanel.setValue("velocity1", vel1);
		controlPanel.setValue("velocity2", vel2);
		update();
		model.setup(mass1, vel1, mass2, vel2, restitution);
		view.setup(model);
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		if (event.getActionCommand().equals("start"))
			start();
		else if (event.getActionCommand().equals("problem1"))
			problemButton(new ElasticCollisionProblem(this));
		else if (event.getActionCommand().equals("problem2"))
			problemButton(new InelasticCollisionProblem(this));
		else if (event.getActionCommand().equals("equations"))
			Problem.popupFrame(EQUATIONS_TITLE, EQUATIONS_TEXT);
		else if (event.getActionCommand().equals("help"))
			JOptionPane.showMessageDialog(this, HELP_TEXT, HELP_TITLE, JOptionPane.PLAIN_MESSAGE);
	}
	
	protected void problemButton(Problem problem) {
		Frame frame = Problem.popupFrame(EQUATIONS_TITLE, EQUATIONS_TEXT);
		problem.show(null);
		frame.dispose();
		problem.setup();
	}
	
	protected void start() {
		double m1 = controlPanel.getValue("mass1");
		double m2 = controlPanel.getValue("mass2");
		double v1 = controlPanel.getValue("velocity1");
		double v2 = controlPanel.getValue("velocity2");
		double restitution = controlPanel.getValue("restitution");
		model.setup(m1, v1, m2, v2, restitution);
		view.setup(model);
		model.start();
	}

	public Component getComponent() {
		return this;
	}

	public void valueChanged(String field, double value) {
		double mass1 = controlPanel.getValue("mass1");
		double mass2 = controlPanel.getValue("mass2");
		double vel1 = controlPanel.getValue("velocity1");
		double vel2 = controlPanel.getValue("velocity2");
		double momentum1 = controlPanel.getValue("momentum1");
		double momentum2 = controlPanel.getValue("momentum2");
		
		if (field == "mass1" || field == "velocity1") {
			momentum1 = mass1 * vel1;
			controlPanel.setValue("momentum1", momentum1);
		}
		else if (field == "mass2" || field == "velocity2") {
			momentum2 = mass2 * vel2;
			controlPanel.setValue("momentum2", momentum2);
		}
		else if (field == "momentum1") {
			vel1 = momentum1 / mass1;
			controlPanel.setValue("velocity1", vel1);
		}
		else if (field == "momentum2") {
			vel2 = momentum2 / mass2;
			controlPanel.setValue("velocity2", vel2);
		}
		
		model.setup(mass1, vel1, mass2, vel2, controlPanel.getValue("restitution"));
		view.setup(model);
	}
}

class Ball {
	
	public double SCALE = 0.1;
	
	public double mass;
	public double radius;
	public double initX; // The initial location
	public double velocity; // The initial velocity
	public double momentum; // The initial momentum
	public double energy; // The initial energy
	
	public double x; // The current location
	
	public Ball(double mass, double x, double velocity) {
		this.mass = mass;
		this.radius = Math.cbrt(this.mass) * SCALE;
		this.initX = x;
		this.velocity = velocity;
		this.momentum = mass * velocity;
		this.energy = mass * velocity * velocity / 2.0;
		
		this.x = initX;
	}
	
	public Point getLocation() {
		return new Point(x, 0.0);
	}
}


class CollisionModel extends Model {
	
	protected Timer timer;
	
	protected Ball ball1;
	protected Ball ball2;
	
	protected double time_coll; // The time of collision
	protected double time_final; // The total running time
	
	protected double restitution; // The coefficient of restitution, which defines how elastic a collision is.
	
	protected int time; // The current time, in milliseconds.
	protected final int PERIOD = 10; // How often to update, in milliseconds.
	
	/**
	 * Have objects 1 and 2
	 * masses: m1, m2
	 * velocities: m1, m2 (positive means to the right, negative to the left)
	 * initial locations: x1, x2
	 * before collision:
	 *     f1(t) = x1 + v1*t
	 *     f2(t) = x2 + v2*t
	 * after collision:
	 *     f1(t) = (m1 - m2)/(m1 + m2)*v1 + 2*m2/(m1 + m2)*v2
	 *     f2(t) = 2*m1/(m1 + m2)*v1 + (m2 - m1)/(m1 + m2)*v2
	 *     
	 * d(t) = |f1(t) - f2(t)|
	 * time until collision = min t>=0 : d(t) = r1 + r2
	 *   (if the objects are moving away from each other, 'time until collision' may be undefined)
	 * total simulation time = 2 * time until collision
	 * 
	 * Now fix x1=0 and x2=L.
	 * The leftmost point is min{f1(0), f1(final)} - r1.
	 * The rightmost point is max{f2(0), f2(final)} + r2.
	 */
	
	public void setup(double mass1, double velocity1, double mass2, double velocity2, double restitution) {
		if (timer != null) {
			timer.cancel();
			timer = null;
		}
		
		ball1 = new Ball(mass1, 0, velocity1);
		ball2 =  new Ball(mass2, CollisionApplet.DISTANCE, velocity2);
		this.restitution = restitution;
		
		this.time = 0;
		
		// Possible collision times:
		double radiusDist = ball1.radius + ball2.radius;
		double time_coll_1 = (radiusDist - ball2.x + ball1.x)/(ball2.velocity - ball1.velocity);
		double time_coll_2 = (radiusDist - ball1.x + ball2.x)/(ball1.velocity - ball2.velocity);
		time_coll = Math.min(time_coll_1, time_coll_2);
		
		// Will they ever collide?
		if (time_coll <= 0.0 || Double.isInfinite(time_coll))
			return;
		
		time_final = 1.5*time_coll;
		
		draw();
	}
	
	public void start() {
		if (timer != null) {
			timer.cancel();
			timer = null;
		}
		
		// Start the timer.
		timer = new Timer();
		timer.scheduleAtFixedRate(new RunTask(), 0, PERIOD);
	}
	
	protected double getLocation(Ball ball, double t) {
		if (t <= time_coll)
			return ball.initX + ball.velocity*t;
		else
			return ball.initX + ball.velocity*time_coll +
				getFinalVelocity(ball)*(t - time_coll);
	}
	
	class RunTask extends TimerTask {
		public void run() {
			time += PERIOD;
			double time_secs = time/1000.0;
			ball1.x = getLocation(ball1, time_secs);
			ball2.x = getLocation(ball2, time_secs);
			draw();
		}
	}
	
	// For view:
	
	public Ball getBall1() {
		return ball1;
	}
	
	public Ball getBall2() {
		return ball2;
	}
	
	// Get the left and right boundaries
	public double getLeft() {
		return Math.min(getLocation(ball1, 0.0), getLocation(ball1, time_final));
	}
	
	public double getRight() {
		return Math.max(getLocation(ball2, 0.0), getLocation(ball2, time_final));
	}
	
	// For controller:
	
	public double getFinalVelocity(Ball ball) {
		double m1 = ball1.mass;
		double m2 = ball2.mass;
		double v1 = ball1.velocity;
		double v2 = ball2.velocity;
		
		if (ball == ball1)
			return collision(restitution, m1, m2, v1, v2);
		else
			return collision(restitution, m2, m1, v2, v1);
	}
	
	// Calculates the final velocity of ball1 after a potentially inelastic collision.
	public static double collision(double r, double m1, double m2, double v1, double v2) {
		return ((r + 1)*m2*v2 + v1*(m1 - r*m2))/(m1 + m2);
	}
}


@SuppressWarnings("serial")
class CollisionView extends View {
	
	public static Color COLOR_A = new Color(0, 0, 100);
	public static Color COLOR_B = new Color(0, 100, 0);
	
	protected double DISTANCE = CollisionApplet.DISTANCE;
	
	protected Graph graph;
	protected AxisPlot axisPlot;
	protected PointPlot plot1;
	protected PointPlot plot2;
	
	public CollisionView() {
		setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
		graph = new Graph(new Dimension(800, 400), this);
		axisPlot = new AxisPlot(graph);
		axisPlot.setRegion(new Region(0, -DISTANCE/4, DISTANCE, DISTANCE/4));
		axisPlot.toggleScale(true);
		axisPlot.toggleNumbers(true);
		axisPlot.toggleYAxis(false);
		axisPlot.setXUnit("m");
		plot1 = new PointPlot(graph, COLOR_A, null);
		plot2 = new PointPlot(graph, COLOR_B, null);
	}
	
	public void setup(CollisionModel model) {
		double left = model.getLeft();
		double right = model.getRight();
		axisPlot.setRegion(new Region(left, -1, right, 1));
		
		// Set the radii of the balls.
		// This is ugly because x-distances and y-distances might be different on the screen.
		Point origin = new Point(0, 0);
		Point radius1 = new Point(model.getBall1().radius, 0);
		Point radius2 = new Point(model.getBall2().radius, 0);
		CoordinateSystem space = graph.getCoordinateSpace();
		plot1.setRadius(space.screenDistance(origin, radius1));
		plot2.setRadius(space.screenDistance(origin, radius2));
		
		draw(model);
	}
	
	@Override
	public void draw(Model m) {
		CollisionModel model = (CollisionModel) m;
		plot1.move(new Point(model.getBall1().getLocation()));
		plot2.move(new Point(model.getBall2().getLocation()));
		graph.draw();
	}

	public Component getComponent() {
		return graph;
	}
}

class ElasticCollisionProblem extends Problem {
	private static String TITLE = "Elastic Collision";
	private static String PROBLEM_TEXT =
		"Two perfectly elastic balls hurdle towards each other in outer space.\n" +
		"The first ball has a mass of %d kg and a velocity of %.2f m/s.\n" +
		"The second ball has a mass of %d kg and a velocity of %.2f m/s.\n\n" +
		"What is the velocity of the second ball after the collision (in m/s)?";
	
	private CollisionController controller;
	private int mass1;
	private int mass2;
	private double vel1;
	private double vel2;
	private double correctAnswer;
	
	public ElasticCollisionProblem(CollisionController controller) {
		this.controller = controller;
		mass1 = randInt(5, 15, 1);
		mass2 = randInt(5, 15, 1);
		vel1 = randDouble(0.5, 2.0, 0.1);
		vel2 = randDouble(-2.0, -0.5, 0.1);
		correctAnswer = CollisionModel.collision(1.0, mass2, mass1, vel2, vel1);
	}
	
	protected Result display(Component parent) {
		String problemText = String.format(PROBLEM_TEXT, mass1, vel1, mass2, vel2);
		double answer = popup(parent, problemText, TITLE);
		return checkAnswer(answer, correctAnswer);
	}

	public void setup() {
		controller.setup(mass1, mass2, vel1, vel2, 1.0);
	}
}

class InelasticCollisionProblem extends Problem {
	private static String TITLE = "Inelastic Collision";
	private static String PROBLEM_TEXT =
		"Two inelastic balls hurdle towards each other in outer space.\n" +
		"The first ball has a mass of %d kg and a velocity of %.2f m/s.\n" +
		"The second ball has a mass of %d kg and a velocity of %.2f m/s.\n" +
		"The coefficient of restitution between the two balls is %.2f.\n\n" +
		"What is the total kinetic energy, in Joules, of the system after\n" +
		"the collision?";
	
	private CollisionController controller;
	private int mass1;
	private int mass2;
	private double vel1;
	private double vel2;
	private double restitution;
	private double correctAnswer;
	
	public InelasticCollisionProblem(CollisionController controller) {
		this.controller = controller;
		mass1 = randInt(5, 15, 1);
		mass2 = randInt(5, 15, 1);
		vel1 = randDouble(1.0, 2.0, 0.1);
		vel2 = randDouble(-2.0, -1.0, 0.1);
		restitution = randDouble(0.3, 0.8, 0.1);
		double vel1f = CollisionModel.collision(restitution, mass1, mass2, vel1, vel2);
		double vel2f = CollisionModel.collision(restitution, mass2, mass1, vel2, vel1);
		correctAnswer = mass1 * vel1f * vel1f / 2.0 + mass2 * vel2f * vel2f / 2.0;
	}
	
	protected Result display(Component parent) {
		String problemText = String.format(PROBLEM_TEXT, mass1, vel1, mass2, vel2, restitution);
		double answer = popup(parent, problemText, TITLE);
		return checkAnswer(answer, correctAnswer);
	}

	public void setup() {
		controller.setup(mass1, mass2, vel1, vel2, restitution);
	}
}