package phyApps;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Timer;
import java.util.TimerTask;

import phyApps.common.ControlPanel;
import phyApps.common.mvc.Controller;
import phyApps.common.mvc.Model;
import phyApps.common.mvc.View;
import phyApps.common.mvc.PhyApplet;
import phyApps.common.graph.AxisPlot;
import phyApps.common.graph.CurvePlot;
import phyApps.common.graph.Graph;
import phyApps.common.graph.Plot;
import phyApps.common.graph.PointPlot;
import phyApps.common.graph.Region;
import phyApps.common.graph.Point;

@SuppressWarnings("serial")
public class Hooke extends PhyApplet {
	
	public void init() {
		setupMVC(new HookeModel(), new HookeView(), new HookeController());
	}
	
	public static void main(String[] args) {
		PhyApplet.runApplet(new Hooke(), "Hooke's Law");
	}
}


/******************************************** Controller ***************************************************/
@SuppressWarnings("serial")
class HookeController extends Controller implements ActionListener {

	private HookeModel model;
	private ControlPanel controlPanel;

	public void setup(View view, Model model) {
		this.model = (HookeModel) model;
		controlPanel = new ControlPanel();
		controlPanel.addButton("start", "Start", this);
		controlPanel.addButton("stop", "Stop", this);
		controlPanel.addLabel("mass", "Mass m:", "kg");
		controlPanel.addLabel("k", "Spring constant k:", "N/m");
		controlPanel.addLabel("f", "Frequency f:", "Hz");
		controlPanel.setValue("f", this.model.frequency);
		controlPanel.setValue("mass", this.model.weight);
		controlPanel.setValue("k", this.model.k);
		
		this.add(controlPanel);
	}
	
	public Component getComponent(){
		return this;
	}
		
	public void actionPerformed(ActionEvent e) {
		String command = e.getActionCommand();
		if (command == "start")
			model.start();
		else if (command == "stop")
			model.stop();
	}

	public void update() {
		// Nothing to be done.
	}
}

/********************************************* Model *******************************************************/
class HookeModel extends Model {

	private Timer timer;
	private int PERIOD = 100; // The period of the timer.
	private double time = 0; // The current time, measured in 10 ticks of the timer.
	public double weight = 10.0;
	public double k = 395; //spring constant
	public double frequency = Math.sqrt(k / weight) / (2 * Math.PI);

	public HookeModel() {
		time = 0;
	}

	public void start() {
		if (timer != null)
			timer.cancel();
		time = 0;
		timer = new Timer();
		timer.scheduleAtFixedRate(new HookeTimer(), (int) time, PERIOD);
	}

	public void stop() {
		if (timer != null)
			timer.cancel();
		time = 0;
		draw();
	}
	
	public double getTime(){
		return time;
	}
	
	public double getWeight(){
		return weight;
	}
	
	public double getLength() {
		return 9.8*weight/k;
	}
	
	public void updateWeight(double w){
		weight = w;
		draw();
	}
	
	class HookeTimer extends TimerTask {
		// This method is called on every tick of the timer.
		public void run() {
			if (time < 4.0*Math.PI)
				time += PERIOD/1000.0;
			else
				time = 0;
			draw();
		}
	}
}


@SuppressWarnings("serial")
class HookeView extends View {

	private HookePlot plot;
	private Graph springGraph;

	public HookeView() {
		springGraph = new Graph(new Dimension(700, 300));
		plot = new HookePlot(springGraph);
	}
	
	public Component getComponent() {
		return springGraph;
	}

	public void draw(Model model) {
		HookeModel m = (HookeModel) model;
		plot.setAmplitude(m.getLength()/2.0);
		plot.setTime(m.getTime());
		springGraph.draw();
	}
}

/********************************************* Plot ********************************************************/
class HookePlot extends Plot {

	private static double MAX_AMP = 0.25; //10
	private static double COMPRESSED_LEN = 0.01; //2
	private static double WIDTH = 4.0;
	private static int WEIGHT_SIZE = 15;
	private double amplitude;
	private double time;
	private CurvePlot sine; //plot of the blue sine wave
	private CurvePlot spring; //plot of the oscillating spring
	private CurvePlot length; //plot of the displacement bar
	private PointPlot follower; //plot of the point that follows the sine wave
	private PointPlot weight; //plot of the weight hanging on the spring
	
	private double v = MAX_AMP/12.0; //coil height 0.75
	private double w = 0.75; //coil width
	private int n = 10; //number of coils
	private double L = 0.0; //spring length
	
	public HookePlot(Graph graph) {
		super(graph);

		AxisPlot axisPlot = new AxisPlot(graph);
		axisPlot.toggleNumbers(true);
		axisPlot.toggleScale(true);
		sine  = new CurvePlot(graph, Color.blue);
		spring = new CurvePlot(graph, new Color(50, 50, 50));
		length = new CurvePlot(graph, Color.black);
		follower = new PointPlot(graph, Color.red);
		weight = new PointPlot(graph, new Color(100, 100, 100));
		
		follower.setRegion(null);
		spring.setRegion(null);
		weight.setRadius(WEIGHT_SIZE);
	}
	
	public void setAmplitude(double amplitude) {
		this.amplitude = amplitude;
	}

	public Region determineRegion() {
		return new Region(-WIDTH, -MAX_AMP, 4*Math.PI, MAX_AMP + COMPRESSED_LEN);
	}

	public boolean isEmpty() {
		return false;
	}
	
	public void setTime(double t){
		time = t;
		spring.clear();
		sine.clear();
		length.clear();
		weight.clear();
		
		L = MAX_AMP - amplitude*Math.sin(time + Math.PI/2.0);

		double springCenter = -WIDTH/2.0;
		
		// Draw the spring.
		for (double s=0; s<=(4*n)-2; s+=0.1) {
			spring.add(new Point(springCenter + w*Math.sin((Math.PI/2)*s),
					MAX_AMP -((L-v)/((4*n)-2))*s + v*Math.cos((Math.PI/2)*s)));
		}
		Point weightCenter = new Point(springCenter, MAX_AMP - L);
		weightCenter = graph.getCoordinateSpace().translate(weightCenter, 0, WEIGHT_SIZE);
		weight.move(weightCenter);
		
		// Draw the amplitude plot.
		for (double j=0; j<=time; j+=0.1) {
			sine.add(new Point(j, amplitude*Math.sin(j + Math.PI/2.0)));
		}
		follower.move(new Point(time, MAX_AMP - L));
	}

	public void draw(Graphics g) {
	}
}
