package phyApps;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Timer;
import java.util.TimerTask;

import phyApps.common.ControlPanel;
import phyApps.common.ControlPanelListener;
import phyApps.common.mvc.Controller;
import phyApps.common.mvc.Model;
import phyApps.common.mvc.View;
import phyApps.common.mvc.PhyApplet;
import phyApps.common.graph.ArrowPlot;
import phyApps.common.graph.AxisPlot;
import phyApps.common.graph.CurvePlot;
import phyApps.common.graph.Graph;
import phyApps.common.graph.Plot;
import phyApps.common.graph.Region;
import phyApps.common.graph.Point;


@SuppressWarnings("serial")
public class Rotation extends PhyApplet {

	public void init() {
		setupMVC(new RotationModel(), new RotationView(), new RotationController());
	}
	
	public static void main(String[] args) {
		PhyApplet.runApplet(new Rotation(), "Rotational Motion");
	}
}
	
//Applet can use setLayout method from Container
//default window size in "run configuration"
/******************************************** Controller ***************************************************/
@SuppressWarnings("serial")
class RotationController extends Controller implements ActionListener, ControlPanelListener {

	private RotationModel model;
	private ControlPanel panel;

	public Component getComponent() {
		return panel;
	}

	public void actionPerformed(ActionEvent e) {
		String command = e.getActionCommand();
		if (command == "start") {
			model.start();
		}
		else if (command == "stop") {
			model.stop();
		}
	}

	@Override
	public void setup(View view, Model m) {
		model = (RotationModel) m;
		
		panel = new ControlPanel();
		
		panel.addScrollbar("radius", "Radius (m)", RotationModel.DEFAULT_RADIUS, 1, 20, 1);
		panel.addScrollbar("frequency", "Frequency (Hz)", 1.0/RotationModel.DEFAULT_PERIOD, 0.05, 0.50, 0.01);
		panel.addScrollbar("period", "Period (s)", RotationModel.DEFAULT_PERIOD, 2, 20, 1);
		
		panel.addListener("radius", this);
		panel.addListener("frequency", this);
		panel.addListener("period", this);

		panel.addButton("start", "Start", this);
		panel.addButton("stop", "Stop", this);
	}

	@Override
	public void update() {
		// do nothing
	}

	@Override
	public void valueChanged(String field, double value) {
		if (field == "radius") {
			model.setRadius(value);
		}
		else if (field == "frequency") {
			double period = 1.0/value;
			model.setPeriod(period);
			panel.setValue("period", period);
		}
		else if (field == "period") {
			double frequency = 1.0/value;
			model.setPeriod(value);
			panel.setValue("frequency", frequency);
		}
	}
}

/********************************************* Model *******************************************************/
class RotationModel extends Model {

	public static double DEFAULT_RADIUS = 10.0;
	public static double DEFAULT_PERIOD = 10.0;
	
	private static int TIMER_PERIOD = 40; // The period of the timer, in ms.
	
	private Timer timer;
	private double time; // The current time, measured in seconds.
	private double period; // The period of the rotating arrow, in seconds
	private double radius; // The radius of the circle, in abstract coordinates
	private double angle; // The current angle of the arrow, in radians.
	
	public RotationModel() {
		time = 0.0;
		angle = 0.0;
		period = DEFAULT_PERIOD;
		radius = DEFAULT_RADIUS;
	}

	public void start() {
		if (timer != null)
			timer.cancel();
		timer = new Timer();
		timer.scheduleAtFixedRate(new Task(), 0, TIMER_PERIOD);
	}

	public void stop() {
		if (timer != null)
			timer.cancel();
		draw();
	}
	
	class Task extends TimerTask {
		// This method is called on every tick of the timer.
		public void run() {
			time += TIMER_PERIOD/1000.0;
			angle += 2.0*Math.PI*TIMER_PERIOD/1000.0/period;
			draw();
		}
	}
	
	public double getAngle() {
		return angle;
	}
	
	public void setPeriod(double p){
		period = p;
		draw();
	}
	
	public double getRadius(){
		return radius;
	}
	
	public void setRadius(double r){
		radius = r;
		draw();
	}
}

/********************************************* View ********************************************************/
@SuppressWarnings("serial")
class RotationView extends View {

	private RotationPlot plot;
	private Graph graph;

	public RotationView() {
		graph = new Graph(new Dimension(600, 600));
		plot = new RotationPlot(graph, RotationModel.DEFAULT_RADIUS);
	}
	
	public Component getComponent() {
		return graph;
	}

	@Override
	public void draw(Model m) {
		RotationModel model = (RotationModel) m;
		plot.setRadius(model.getRadius());
		plot.setAngle(model.getAngle());
		graph.draw();
	}
}

/********************************************* Plot ********************************************************/
class RotationPlot extends Plot {

	private double radius;
	private CurvePlot circle; //plot of the stationary circle, radius variable from 0-20
	private ArrowPlot arrow; //plot of the rotating arrow
	private AxisPlot axes;
	
	public RotationPlot(Graph graph, double r) {
		super(graph);
		radius = r;
		
		axes = new AxisPlot(graph);
		axes.toggleScale(true);
		axes.toggleNumbers(true);
		axes.setRegion(new Region (-20, -20, 20, 20));
		circle = new CurvePlot(graph, Color.blue);
		arrow = new ArrowPlot(graph, Color.red);
		
		for (double i=0; i<=2*Math.PI; i+=0.1){
			circle.add(new Point(radius*Math.cos(i), radius*Math.sin(i)));
		}
	}
	
	public Region determineRegion() {
		return new Region(-radius, radius, -radius, radius);
	}

	public void draw(Graphics g) {
	}

	public boolean isEmpty() {
		return false;
	}
	
	public void setRadius(double r){
		radius = r;
		circle.clear();
		
		for (double i=0; i<=2*Math.PI+0.05; i+=0.05){
			circle.add(new Point(radius*Math.cos(i), radius*Math.sin(i)));
		}
	}
	
	public void setAngle(double angle) {
		arrow.move(new Point(0, 0), new Point(radius*Math.cos(angle), radius*Math.sin(angle)));
	}
}
