package phyApps;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import phyApps.common.ControlPanel;
import phyApps.common.graph.ArrowPlot;
import phyApps.common.graph.AxisPlot;
import phyApps.common.graph.CoordinateSystem;
import phyApps.common.graph.Graph;
import phyApps.common.graph.Point;
import phyApps.common.graph.Region;
import phyApps.common.mvc.Controller;
import phyApps.common.mvc.Model;
import phyApps.common.mvc.View;
import phyApps.common.mvc.PhyApplet;

@SuppressWarnings("serial")
public class VectorApplet extends PhyApplet {

	public void init() {
		setupMVC(new VectorModel(), new VectorView(), new VectorController());
	}
	
	public static void main(String[] args) {
		PhyApplet.runApplet(new VectorApplet(), "2D Vector Addition");
	}
}

@SuppressWarnings("serial")
class VectorView extends View {
	
	private Graph graph;
	private AxisPlot axisPlot;
	private ArrowPlot vectorX;
	private ArrowPlot vectorY;
	private ArrowPlot vectorX2;
	private ArrowPlot vectorY2;
	private ArrowPlot sumVector;
	
	public static Color GRID_COLOR = new Color(80, 80, 80);
	public static Color X_COLOR = new Color(25, 25, 255);
	public static Color Y_COLOR = new Color(25, 155, 25);
	public static Color X2_COLOR = new Color(120, 120, 220);
	public static Color Y2_COLOR = new Color(120, 220, 120);
	public static Color SUM_COLOR = new Color(230, 25, 25);
	public static Color DOT_PRODUCT_COLOR = Color.BLACK;
	
	private static double ARROW_THICKNESS = 2.25;
	
	public VectorView() {
		setLayout(new FlowLayout());
		graph = new Graph(new Dimension(600, 600));
		
		axisPlot = new AxisPlot(graph, GRID_COLOR);
		axisPlot.toggleScale(true);
		axisPlot.toggleNumbers(true);
		axisPlot.toggleDenseGrid(true, 1.0, 1.0);

		vectorX2 = new ArrowPlot(graph, X2_COLOR, null);
		vectorY2 = new ArrowPlot(graph, Y2_COLOR, null);
		vectorX = new ArrowPlot(graph, X_COLOR, null);
		vectorY = new ArrowPlot(graph, Y_COLOR, null);
		sumVector = new ArrowPlot(graph, SUM_COLOR, null);
		vectorX.thickness = ARROW_THICKNESS;
		vectorY.thickness = ARROW_THICKNESS;
		vectorX2.thickness = ARROW_THICKNESS;
		vectorY2.thickness = ARROW_THICKNESS;
		sumVector.thickness = ARROW_THICKNESS;
		
		add(graph);
	}
	
	public void draw(Model m) {
		VectorModel model = (VectorModel) m;
		setScale(model.getScale());
		Point X = model.getX();
		Point Y = model.getY();
		Point S = model.getSum();
		
		if (X == null && Y == null) {
			vectorX.clear();
			vectorY.clear();
			vectorX2.clear();
			vectorY2.clear();
			sumVector.clear();
		}
		else if (X != null && Y == null) {
			vectorX.move(new Point(0, 0), X);
			vectorY.clear();
			vectorX2.clear();
			vectorY2.clear();
			sumVector.clear();
		}
		else if (X != null && Y != null) {
			vectorX.move(new Point(0, 0), X);
			vectorY.move(new Point(0, 0), Y);
			vectorX2.move(Y, S);
			vectorY2.move(X, S);
			sumVector.move(new Point(0, 0), S);
		}
		
		graph.draw();
	}
	
	public void setScale(double scale) {
		axisPlot.setRegion(new Region(-scale, -scale, scale, scale));
	}
	
	// For the controller
	
	public CoordinateSystem getCoordinateSpace() {
		return graph.getCoordinateSpace();
	}
	
	@Override
	public void addMouseListener(MouseListener listener) {
		graph.addMouseListener(listener);
	}
	
	@Override
	public void addMouseMotionListener(MouseMotionListener listener) {
		graph.addMouseMotionListener(listener);
	}

	public Component getComponent() {
		return graph;
	}
}


class VectorModel extends Model {
	
	private double scale = 10.0; // The default size of the view. (-MIN_SIZE < x, y < MIN_SIZE)
	private Point a;
	private Point b;
	
	// For view:
	
	public Point getX() {
		return a;
	}
	
	public Point getY() {
		return b;
	}
	
	public Point getSum() {
		if (a == null || b == null)
			return null;
		return Point.sum(a, b);
	}
	
	public Point getDotProduct() {
		if (a == null || b == null)
			return null;
		return new Point(a.getX()*b.getX() + a.getY()*b.getY(), 0);
	}
	
	public double getScale() {
		return scale;
	}
	
	// For controller:
	
	public void add(Point p) {
		if (a == null && b == null)
			a = p;
		else if (a != null && b == null)
			b = p;
		else if (a != null && b != null) {
			a = p;
			b = null;
		}
		draw();
	}
	
	public void setX(Point x) {
		a = x;
		draw();
	}
	
	public void setY(Point y) {
		b = y;
		draw();
	}
	
	public void setSum(Point sum) {
		b = new Point(sum.getX() - a.getX(), sum.getY() - a.getY());
		draw();
	}
}

@SuppressWarnings("serial")
class VectorController extends Controller implements MouseListener, MouseMotionListener {

	private VectorView view;
	private VectorModel model;
	private VectorSelection dragged; // The vector being dragged.
	private ControlPanel controlPanel;
	
	private double PRECISION = 10.0;
	
	public VectorController() {
		controlPanel = new ControlPanel();
		
		controlPanel.addLabel("x", "Vector 1");
		controlPanel.addLabel("y", "Vector 2");
		controlPanel.addLabel("sum", "Sum");
		controlPanel.addLabel("product", "Dot Product");
		controlPanel.setColor("x", VectorView.X_COLOR);
		controlPanel.setColor("y", VectorView.Y_COLOR);
		controlPanel.setColor("sum", VectorView.SUM_COLOR);
		controlPanel.setColor("product", VectorView.DOT_PRODUCT_COLOR);
		
		add(controlPanel);
	}
	
	private enum VectorSelection {
		X_VECTOR,
		Y_VECTOR,
		SUM,
		NONE
	}
	
	@Override
	public void setup(View view, Model model) {
		this.view = (VectorView) view;
		this.model = (VectorModel) model;
		view.addMouseListener(this);
		view.addMouseMotionListener(this);
		dragged = VectorSelection.NONE;
		this.view.setScale(this.model.getScale());
	}
	
	@Override
	public void mousePressed(MouseEvent event) {
		Point location = view.getCoordinateSpace().makePoint(event.getX(), event.getY());
		if (near(location, model.getX()))
			dragged = VectorSelection.X_VECTOR;
		else if (near(location, model.getY()))
			dragged = VectorSelection.Y_VECTOR;
		else if (near(location, model.getSum()))
			dragged = VectorSelection.SUM;
		else {
			model.add(getPoint(event));
			dragged = VectorSelection.NONE;
		}
	}
	
	@Override
	public void mouseDragged(MouseEvent event) {
		if (dragged == VectorSelection.X_VECTOR) {
			model.setX(getPoint(event));
		}
		else if (dragged == VectorSelection.Y_VECTOR) {
			model.setY(getPoint(event));
		}
		else if (dragged == VectorSelection.SUM) {
			model.setSum(getPoint(event));
		}
	}
	
	// Translate a mouse click location into abstract coordinates.
	protected Point getPoint(MouseEvent event) {
		int x = event.getX();
		int y = event.getY();
		Point point = view.getCoordinateSpace().makePoint(x, y);
		double m = PRECISION/model.getScale();
		return new Point(Math.round(m*point.getX())/m, Math.round(m*point.getY())/m);
	}
	
	protected boolean near(Point p1, Point p2) {
		if (p1 == null || p2 == null)
			return false;
		else
			return view.getCoordinateSpace().screenDistance(p1, p2) < 25;
	}
	
	public void update() {
		updateLabel("x", model.getX());
		updateLabel("y", model.getY());
		updateLabel("sum", model.getSum());
		
		Point dotProduct = model.getDotProduct();
		if (dotProduct == null)
			controlPanel.setValue("product", "");
		else
			controlPanel.setValue("product", Graph.formatDouble(dotProduct.getX(), model.getScale()/PRECISION));
	}

	private void updateLabel(String field, Point P) {
		if (P == null) {
			controlPanel.setValue(field, "");
		}
		else {
			//int power = (int) Math.round(Math.log10(model.getScale()));
			String xStr = Graph.formatDouble(P.getX(), model.getScale()/PRECISION);
			String yStr = Graph.formatDouble(P.getY(), model.getScale()/PRECISION);
			controlPanel.setValue(field, xStr + "i + " + yStr + "j");
		}
	}

	public Component getComponent() {
		return this;
	}

}

