package phyApps.common;

import java.awt.Button;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Label;
import java.awt.Panel;
import java.awt.Scrollbar;
import java.awt.TextField;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.util.HashMap;
import java.util.Map;

import phyApps.common.graph.Graph;

@SuppressWarnings("serial")
public class ControlPanel extends Panel {

	public static double SCROLLBAR_WIDTH = 1.0/3.0;
	public static int COLUMN_WIDTH = 6;
	public static String LABEL_STRING = "            ";
	
	private Map<String, Field> fields;
	
	public ControlPanel() {
		fields = new HashMap<String, Field>();
		setLayout(new GridBagLayout());
		setBackground(Color.LIGHT_GRAY);
	}
	
	public void clear() {
		fields = new HashMap<String, Field>();
		this.removeAll();
	}

	public void addText(String text) {
		addText(text, Label.LEFT);
	}
	
	public void addText(String text, int alignment) {
		Label label = new Label(text);
		label.setAlignment(alignment);
		addOneComponent(label);
	}
	
	public void addText(String text, int alignment, Font font) {
		Label label = new Label(text);
		label.setAlignment(alignment);
		label.setFont(font);
		addOneComponent(label);
	}
	
	public Button addButton(String name, String description, ActionListener listener) {
		return addButton(name, description, listener, true);
	}
	
	public Button addButton(String name, String description, ActionListener listener, boolean fillWidth) {
		Button button = new Button(description);
		button.setActionCommand(name);
		button.addActionListener(listener);
		addOneComponent(button, fillWidth);
		return button;
	}

	public void addLabel(String name, String description) {
		addLabel(name, description, null);
	}
	
	public void addLabel(String name, String description, String units) {
		fields.put(name, new LabelField(this, description, units));
	}

	public void addTextField(String name, String description, double defaultVal) {
		addTextField(name, description, defaultVal, null);
	}
	
	public void addTextField(String name, String description, double defaultVal, String units) {
		addTextField(name, description, defaultVal, units, Double.NaN, Double.NaN);
	}
	
	public void addTextField(String name, String description, double defaultVal, String units, double min, double max) {
		fields.put(name, new TextFieldField(this, name, description, defaultVal, units, min, max));
	}
	
	public void addScrollbar(String name, String description, double defaultVal, double min, double max, double increment) {
		fields.put(name, new ScrollbarField(this, name, description, defaultVal, min, max, increment));
	}
	
	public double getValue(String name) {
		return fields.get(name).get();
	}
	
	public void setValue(String name, double value) {
		fields.get(name).set(value);
	}
	
	public void setValue(String name, String string) {
		fields.get(name).set(string);
	}
	
	public void setColor(String name, Color color) {
		fields.get(name).setColor(color);
	}

	public void addListener(String name, ControlPanelListener listener) {
		fields.get(name).addListener(listener);
	}
	
	protected void addOneComponent(Component component) {
		addOneComponent(component, true);
	}
	
	protected void addOneComponent(Component component, boolean fillWidth) {
		GridBagConstraints constraints = new GridBagConstraints();
		constraints.gridy = GridBagConstraints.RELATIVE;
		constraints.gridx = 0;
		constraints.gridwidth = 2;
		if (fillWidth)
			constraints.fill = GridBagConstraints.HORIZONTAL;
		else
			constraints.fill = GridBagConstraints.NONE;

		add(component, constraints);
	}
	
	protected void addTwoComponents(Component comp1, Component comp2) {
		GridBagConstraints constraints = new GridBagConstraints();
		constraints.gridy = GridBagConstraints.RELATIVE;
		constraints.fill = GridBagConstraints.HORIZONTAL;
		
		constraints.gridx = 0;
		add(comp1, constraints);
		
		constraints.gridx = 1;
		add(comp2, constraints);
	}
	
	public static double readDouble(String text) {
		double result;
		try {
			result = Double.parseDouble(text);
		}
		catch (NumberFormatException e) {
			result = 0.0;
		}
		return result;
	}
	
	public static double readDouble(TextField field) {
		double result;
		try {
			result = Double.parseDouble(field.getText());
		}
		catch (NumberFormatException e) {
			result = 0.0;
		}
		return result;
	}
	
	public static double readDouble(Label label) {
		double result;
		try {
			result = Double.parseDouble(label.getText());
		}
		catch (NumberFormatException e) {
			result = 0.0;
		}
		return result;
	}
	
	public static int readInteger(TextField field) {
		int result;
		try {
			result = Integer.parseInt(field.getText());
		}
		catch (NumberFormatException e) {
			result = 0;
		}
		return result;
	}
	
	public static String formatDouble(double value, String units) {
		if (units == null || units == "")
			units = "";
		else
			units = " " + units;
		
		double scale = Math.abs(value) / 100.0;
		if (scale < 0.000001 && scale > -0.000001)
			value = 0.0;
		
		return Graph.formatDouble(value, scale, units);
	}

	/**
	 * Enable or disable a field.
	 * @param string - the field name.
	 * @param b - if true, the field will be enabled.
	 */
	public void setEnabled(String string, boolean b) {
		Field field = fields.get(string);
		field.setEnabled(b);
	}
}

interface Field {
	public void set(double value);
	public void set(String string);
	public double get();
	public void addListener(ControlPanelListener listener);
	public void setColor(Color color);
	public void setEnabled(boolean b);
}

class LabelField implements Field {
	private Label label;
	private Label description;
	private String units;
	private double value;
	
	public LabelField(ControlPanel panel, String description, String units) {
		this.units = units;
		this.description = new Label(description);
		label = new Label(ControlPanel.LABEL_STRING);
		panel.addTwoComponents(this.description, this.label);
	}
	
	public void addListener(ControlPanelListener listener) {
		return;
	}

	public double get() {
		return value;
	}

	public void set(double value) {
		this.value = value;
		label.setText(ControlPanel.formatDouble(value, units));
	}
	
	public void set(String string) {
		this.value = 0;
		label.setText(string);
	}

	public void setColor(Color color) {
		this.description.setForeground(color);
	}

	public void setEnabled(boolean b) {
		this.label.setEnabled(b);
	}
}

class TextFieldField implements Field, ActionListener {
	private Label description;
	private TextField field;
	private String name;
	private ControlPanelListener listener;
	private double value;
	private double min;
	private double max;
	
	public TextFieldField(ControlPanel panel, String name, String description, double defaultVal, String units, double min, double max) {
		if (units != null && !units.equals(""))
			description = description + " (" + units + ")";
		this.description = new Label(description);
		this.name = name;
		this.min = min;
		this.max = max;
		field = new TextField(ControlPanel.COLUMN_WIDTH);
		field.addActionListener(this);
		this.set(defaultVal);
		panel.addTwoComponents(this.description, field);
	}

	public void addListener(ControlPanelListener listener) {
		this.listener = listener;
	}

	public double get() {
		return ControlPanel.readDouble(field);
	}

	public void set(double value) {
		field.setText(ControlPanel.formatDouble(value, null));
		this.value = value;
	}

	public void set(String string) {
		field.setText(string);
	}
	
	public void actionPerformed(ActionEvent event) {
		value = ControlPanel.readDouble(field);
		if (!Double.isNaN(min) && value < min)
			set(min);
		if (!Double.isNaN(max) && value > max)
			set(max);
		if (listener != null)
			listener.valueChanged(name, value);
	}

	public void setColor(Color color) {
		this.description.setForeground(color);
	}

	public void setEnabled(boolean b) {
		this.field.setEnabled(b);
	}
}


class ScrollbarField implements Field, AdjustmentListener, ActionListener {
	private Scrollbar scrollbar;
	private String name;
	private TextField text;
	private Label description;
	private ControlPanelListener listener;
	
	private double min;
	private double max;
	private double increment;
	private double length;
	private double value;
	
	public ScrollbarField(ControlPanel panel, String name, String description,
			double defaultVal, double minVal, double maxVal, double increment){
		this.name = name;
		this.description = new Label(description);
		
		length = (maxVal - minVal) / increment;
		this.increment = increment;
		min = minVal;
		max = maxVal;

		double K = ControlPanel.SCROLLBAR_WIDTH;
		int width = (int) Math.round(K / (1.0 - K) * length);
		int scrollbarSize = (int) (width + Math.round(length));
		
		scrollbar = new Scrollbar(Scrollbar.HORIZONTAL, 0, width, 0, scrollbarSize);
		text = new TextField(ControlPanel.COLUMN_WIDTH);
		scrollbar.addAdjustmentListener(this);
		text.addActionListener(this);
		panel.addTwoComponents(this.description, text);
		panel.addOneComponent(scrollbar);
		
		set(defaultVal);
	}

	public void addListener(ControlPanelListener listener) {
		this.listener = listener;
	}

	// something changed : validate, update text & scrollbar, notify listener
	// set : validate, update text & scrollbar
	// get : read text field, validate, update text & scrollbar
	
	public double get() {
		set(ControlPanel.readDouble(text));
		return value;
	}

	public void set(double value) {
		scrollbar.setValue(translate(value));
		text.setText(ControlPanel.formatDouble(translate(scrollbar.getValue()), null));
		this.value = translate(scrollbar.getValue());
	}
	
	public void set(String string) {
		set(ControlPanel.readDouble(string));
	}

	public void adjustmentValueChanged(AdjustmentEvent event) {
		this.value = translate(event.getValue());
		text.setText(ControlPanel.formatDouble(this.value, null));
		if (listener != null)
			listener.valueChanged(name, value);
	}

	public void actionPerformed(ActionEvent event) {
		set(event.getActionCommand());
		if (listener != null)
			listener.valueChanged(name, value);
	}
	
	private int translate(double x) {
		return (int) Math.round((x - min) / (max - min) * length);
	}
	
	private double translate(int x) {
		return (double) x * increment + min;
	}

	public void setColor(Color color) {
		this.description.setForeground(color);
	}

	public void setEnabled(boolean b) {
		this.text.setEnabled(b);
		this.scrollbar.setEnabled(b);
	}
}