package phyApps.common;

import java.awt.Button;
import java.awt.Color;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Label;
import java.awt.Panel;
import java.awt.Scrollbar;
import java.awt.TextField;
import java.awt.event.ActionListener;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.TextEvent;
import java.awt.event.TextListener;
import java.util.HashMap;
import java.util.Map;

@SuppressWarnings("serial")
public class SimpleInputArea extends InputArea implements AdjustmentListener, TextListener {
	
	private Map<Scrollbar, TextField> scrollbarToField;
	private Map<TextField, Scrollbar> fieldToScrollbar;
	private Map<Scrollbar, AdjustmentListener> scrollbarToListener;
	
	public SimpleInputArea(Panel panel) {
		scrollbarToField = new HashMap<Scrollbar, TextField>();
		fieldToScrollbar = new HashMap<TextField, Scrollbar>();
		scrollbarToListener = new HashMap<Scrollbar, AdjustmentListener>();
		
		setLayout(new GridBagLayout());
		setBackground(Color.LIGHT_GRAY);
		panel.add(this);
	}
	
	// Methods for adding scroll bars, buttons, text fields, etc.
	
	/**
	 * Add a button that tells the 'listener' when it is pressed.
	 * It will give 'command' as the event's action command.
	 * @param button - the button to add
	 * @param command - actionEvent.getActionCommand() will return this string
	 * @param listener - listener.actionPerformed() will be called when this button is pressed.
	 */
	public void add(Button button, String command, ActionListener listener) {
		button.setActionCommand(command);
		button.addActionListener(listener);
		addOneComponent(button);
	}
	
	public void add(Scrollbar scrollbar, AdjustmentListener listener) {
		add(scrollbar, listener, null);
	}
	
	/**
	 * Add a scrollbar with the given name.
	 * The name will be shown on a label on top of the scrollbar.
	 * Whenever the scrollbar is moved, an AdjustmentEvent will be sent to listener.
	 * Whenever the scrollbar's value is edited by hand, an Adjustment event will be sent to listener.
	 * @param name - a string describing the scroll bar.
	 * @param scrollbar - the scrollbar to be added.
	 */
	public void add(Scrollbar scrollbar, AdjustmentListener listener, String name) {
		TextField field = new TextField();
		field.setText(Integer.toString(scrollbar.getValue()));
		field.addTextListener(this);
		
		if (name == null)
			addOneComponent(field);
		else
			addTwoComponents(new Label(name), field);
		addOneComponent(scrollbar);
		
		scrollbarToField.put(scrollbar, field);
		fieldToScrollbar.put(field, scrollbar);
		scrollbarToListener.put(scrollbar, listener);
		
		scrollbar.addAdjustmentListener(this);
	}
	
	public void add(TextField field) {
		add(field, "");
	}
	
	/**
	 * Add a text field with the given label string.
	 * The label will be shown to the left of the field.
	 */
	public void add(TextField field, String label) {
		addTwoComponents(new Label(label), field);
	}
	
	/**
	 * Add a label.
	 */
	public void add(Label label) {
		addOneComponent(label);
	}
	
	public void add(Label label, String name) {
		addTwoComponents(new Label(name), label);
	}
	
	/**
	 * Safely read an integer from a text field. Defaults to 0.
	 */
	public static int readInt(TextField field) {
		int result;
		try {
			result = Integer.parseInt(field.getText());
		}
		catch (NumberFormatException e) {
			result = 0;
		}
		return result;
	}
	
	/**
	 * Safely read a double from a text field. Defaults to 0.0.
	 */
	public static double readDouble(TextField field) {
		double result;
		try {
			result = Double.parseDouble(field.getText());
		}
		catch (NumberFormatException e) {
			result = 0.0;
		}
		return result;
	}
	
	public static double readDouble(Label label) {
		double result;
		try {
			result = Double.parseDouble(label.getText());
		}
		catch (NumberFormatException e) {
			result = 0.0;
		}
		return result;
	}

	public void adjustmentValueChanged(AdjustmentEvent event) {
		// Set the text to reflect the scrollbar.
		Scrollbar scrollbar = (Scrollbar) event.getSource();
		TextField field = scrollbarToField.get(scrollbar);
		field.setText(Integer.toString(event.getValue()));
		// Give the listener an adjustment event.
		scrollbarToListener.get(scrollbar).adjustmentValueChanged(event);
	}

	public void textValueChanged(TextEvent event) {
		// Set the scrollbar to reflect the text.
		TextField field = (TextField) event.getSource();
		Scrollbar scrollbar = fieldToScrollbar.get(field);
		int value = readInt(field);
		scrollbar.setValue(value);
		// Give the listener an adjustment event.
		AdjustmentEvent adjustmentEvent = new AdjustmentEvent(scrollbar, AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED, AdjustmentEvent.TRACK, value);
		scrollbarToListener.get(scrollbar).adjustmentValueChanged(adjustmentEvent);
	}
	
	/**
	 * Helper methods
	 */
	
	private void addOneComponent(Component component) {
		GridBagConstraints constraints = new GridBagConstraints();
		constraints.gridy = GridBagConstraints.RELATIVE;
		constraints.gridx = 0;
		constraints.gridwidth = 2;
		constraints.fill = GridBagConstraints.HORIZONTAL;
		
		add(component, constraints);
	}
	
	protected void addTwoComponents(Component comp1, Component comp2) {
		GridBagConstraints constraints = new GridBagConstraints();
		constraints.gridy = GridBagConstraints.RELATIVE;
		constraints.fill = GridBagConstraints.HORIZONTAL;
		
		constraints.gridx = 0;
		add(comp1, constraints);
		
		constraints.gridx = 1;
		add(comp2, constraints);
	}
}
