package phyApps.common.graph;

import java.awt.Color;
import java.awt.Graphics;

import javax.swing.text.View;

/**
 * A plot to draw the x and y axes of a graph.
 * It may also show scale marks and label the axes.
 * @author JLP
 */
public class AxisPlot extends Plot {

	public int THICKNESS = 2; // The thickness of the axis lines
	public int TICK_SIZE = 5; // The size of tick marks on the axes
	
	protected boolean showScale = false;
	protected boolean showNumbers = false;
	protected boolean showXAxis = true;
	protected boolean showYAxis = true;
	protected boolean showGrid = false;
	protected boolean showDenseGrid = false;
	protected double denseGridXIncrement;
	protected double denseGridYIncrement;
	protected Color color;
	protected String xLabel = null;
	protected String yLabel = null;
	protected String xUnit = null;
	protected String yUnit = null;
	
	public AxisPlot(Graph graph) {
		this(graph, Color.BLACK);
	}
	
	public AxisPlot(Graph graph, Color color) {
		super(graph);
		this.color = color;
	}
	
	/**
	 * Label the x axis. 
	 * Unimplemented.
	 * @param label - the label, or null for no label.
	 */
	public void labelXAxis(String label) {
		this.xLabel = label;
	}
	
	/**
	 * Label the y axis. 
	 * Unimplemented.
	 * @param label - the label, or null for no label.
	 */
	public void labelYAxis(String label) {
		this.yLabel = label;
	}
	
	/**
	 * Set the units displayed on the x axis.
	 * @param unit
	 */
	public void setXUnit(String unit) {
		this.xUnit = unit;
	}
	
	/**
	 * Set the units displayed on the x axis.
	 * @param unit
	 */
	public void setYUnit(String unit) {
		this.yUnit = unit;
	}
	
	/**
	 * Should tick marks should be drawn on the axes?
	 * @param visible - if true, the scale will be drawn.
	 */
	public void toggleScale(boolean visible) {
		this.showScale = visible;
	}
	
	/**
	 * Should numbers be shown on the axes?
	 * @param visible - if true, the x and y axes will be numbered.
	 */
	public void toggleNumbers(boolean visible) {
		this.showNumbers = visible;
	}
	
	/**
	 * Show the X axis be drawn?
	 * @param visible - drawn if true
	 */
	public void toggleXAxis(boolean visible) {
		this.showXAxis = visible;
	}
	
	/**
	 * Should a grid be drawn?
	 * @param visible - drawn if true
	 */
	public void toggleGrid(boolean visible) {
		this.showGrid = visible;
	}
	
	/**
	 * Should a dense grid be drawn, with user-defined spacing?
	 * @param visible - drawn if true
	 * @param xIncrement - spacing between vertical grid lines
	 * @param yIncrement - spacing between horizontal grid lines
	 */
	public void toggleDenseGrid(boolean visible, double xIncrement, double yIncrement) {
		this.showDenseGrid = visible;
		this.denseGridXIncrement = xIncrement;
		this.denseGridYIncrement = yIncrement;
	}
	
	/**
	 * Show the Y axis be drawn?
	 * @param on - drawn if true
	 */
	public void toggleYAxis(boolean on) {
		this.showYAxis = on;
	}
	
	// For Graph:
	
	public Region determineRegion() {
		// The origin should always be visible.
		return new Region(0, 0, 0, 0);
	}

	public void draw(Graphics g) {
		Region view = graph.getView();
		
		// Draw the axes.
		if (showXAxis)
			graph.drawLine(g, new Point(view.left, 0), new Point(view.right, 0), color, THICKNESS);
		if (showYAxis)
			graph.drawLine(g, new Point(0, view.bottom), new Point(0, view.top), color, THICKNESS);
		
		if (showScale || showNumbers || showGrid)
			drawScale(g);
		
		if (showDenseGrid)
			drawDenseGrid(g);
	}
	
	protected void drawDenseGrid(Graphics g) {
		Region view = graph.getView();
		
		double lowestMark = denseGridXIncrement * Math.ceil(view.left / denseGridXIncrement);
		for (double mark = lowestMark; mark < view.right; mark += denseGridXIncrement) {
			Point bottom = new Point(mark, view.bottom);
			Point top = new Point(mark, view.top);
			graph.drawLine(g, bottom, top, color);
		}
		
		lowestMark = denseGridYIncrement * Math.ceil(view.left / denseGridYIncrement);
		for (double mark = lowestMark; mark < view.top; mark += denseGridYIncrement) {
			Point left = new Point(view.left, mark);
			Point right = new Point(view.right, mark);
			graph.drawLine(g, left, right, color);
		}
	}
	
	protected void drawScale(Graphics g) {
		Region view = graph.getView();
		CoordinateSystem space = graph.getCoordinateSpace();
		
		// Draw the horizontal scale marks.
		if (showXAxis && view.right - view.left != 0) {
			double increment = getScaleIncrement(view.left, view.right, true);
			double lowestMark = increment * Math.ceil(view.left / increment);
			for (double mark = lowestMark; mark < view.right; mark += increment) {
				if (showScale) {
					Point location = new Point(mark, 0);
					Point bottom = space.translate(location, 0, TICK_SIZE);
					Point top = space.translate(location, 0, -TICK_SIZE);
					graph.drawLine(g, bottom, top, color, 1);
				}
				if (showNumbers) {
					Point location = new Point(mark, 0);
					location = space.translate(location, 3, 15);
					String number = Graph.formatDouble(mark, increment, xUnit);
					graph.drawString(g, number, location, null);
				}
				if (showGrid) {
					Point bottom = new Point(mark, view.bottom);
					Point top = new Point(mark, view.top);
					graph.drawLine(g, bottom, top, color);
				}
			}
		}
		
		// Draw the vertical scale marks.
		if (showYAxis && view.top - view.bottom != 0) {
			double increment = getScaleIncrement(view.bottom, view.top, true);
			double lowestMark = increment * Math.ceil(view.bottom / increment);

			for (double mark = lowestMark; mark < view.top; mark += increment) {
				if (showScale) {
					Point location = new Point(0, mark);
					Point left = space.translate(location, -TICK_SIZE, 0);
					Point right = space.translate(location, TICK_SIZE, 0);
					graph.drawLine(g, left, right, color, 1);
				}
				if (showNumbers) {
					Point location = new Point(0, mark);
					// Adjust somewhat for the size of the numbers.
					String number = Graph.formatDouble(mark, increment, yUnit);
					int numberLength = 8*number.length();
					location = space.translate(location, -5 - numberLength, 3);
					graph.drawString(g, number, location, null);
				}
				if (showGrid) {
					Point left = new Point(view.left, mark);
					Point right = new Point(view.right, mark);
					graph.drawLine(g, left, right, color);
				}
			}

		}
	}
	
	protected void drawNumbers(Graphics g) {
		Region view = graph.getView();
		CoordinateSystem space = graph.getCoordinateSpace();
		
		// Add numbers to the x-axis
		if (showXAxis && view.right - view.left != 0) {
			double increment = getScaleIncrement(view.left, view.right, true);
			double lowestNum = increment * Math.ceil(view.left / increment);
			for (double mark = lowestNum; mark < view.right; mark += increment) {
				Point location = new Point(mark, 0);
				location = space.translate(location, 3, 15);
				String number = Graph.formatDouble(mark, increment, xUnit);
				graph.drawString(g, number, location, null);
			}
		}
		
		// Add numbers to the y-axis
		if (showYAxis && view.top - view.bottom != 0) {
			double increment = getScaleIncrement(view.bottom, view.top, true);
			double lowestNum = increment * Math.ceil(view.bottom / increment);
			for (double mark = lowestNum; mark < view.top; mark += increment) {
				Point location = new Point(0, mark);
				// Adjust somewhat for the size of the numbers.
				String number = Graph.formatDouble(mark, increment, yUnit);
				int numberLength = 8*number.length();
				location = space.translate(location, -5 - numberLength, 3);
				graph.drawString(g, number, location, null);
			}
		}
	}
	
	public boolean isEmpty() {
		return false;
	}
	
	// Helper functions:
	
	private double getScaleIncrement(double lower, double upper, boolean small) {
		// Pull out the largest possible power of ten.
		double power = Math.pow(10, Math.floor((Math.log10(upper - lower))));
		// Round down slightly for when (upper - lower) is a multiple of power.
		power = power - 1e-12;
		// Find the remainder.
		double mantissa = (upper - lower)/power;
		// Choose an appropriate scale.
		// If small == true, there will be between 2 and 5 marks on the scale.
		// If small == false, there will be between 4 and 10 marks.
		if (mantissa < 2.0)
			return small ? 0.2*power : 0.1*power;
		else if (mantissa < 5.0)
			return small ? 0.5*power : 0.2*power;
		else
			return small ? 1.0*power : 0.5*power;
	}
}
