package phyApps.common.graph;

/**
 * A coordinate system keeps track of both the abstract coordinates and screen coordinates.
 * Screen coordinates are the actual pixel locations where things will be drawn. They are integer coordinates.
 * Abstract coordinates are doubles.
 * In screen coordinates, y=0 is at the top of the screen, while in abstract coordinates y=0 is at the bottom.
 * @author JLP
 */
public class CoordinateSystem {

	public Region viewRegion;
	public Region screenRegion;
	
	public CoordinateSystem(Region viewDimensions, Region screenDimensions) {
		viewRegion = viewDimensions;
		screenRegion = screenDimensions;
	}
	
	/**
	 * Get the x location of the point on the screen.
	 * @param point - the point, in abstract coordinates
	 * @return - the number of pixels from the left of the window
	 */
	public int getX(Point point) {
		return changeCoordinates(point.getX(), viewRegion.left, viewRegion.right,
				(int) screenRegion.left, (int) screenRegion.right);
	}
	
	/**
	 * Get the y location of the point on the screen.
	 * @param point - the point, in abstract coordinates
	 * @return - the number of pixels from the top of the window
	 */
	public int getY(Point point) {
		double y = viewRegion.bottom + viewRegion.top - point.getY();
		return changeCoordinates(y, viewRegion.bottom, viewRegion.top,
				(int) screenRegion.bottom, (int) screenRegion.top);
	}
	
	/**
	 * Construct a new point using screen coordinates.
	 * @param x - the x location of the point, in pixels
	 * @param y - the y location of the point, in pixels
	 * @return - The constructed point, in abstract coordinates.
	 */
	public Point makePoint(int x, int y) {
		double viewX = changeCoordinates(x, (int) screenRegion.left, (int) screenRegion.right, viewRegion.left, viewRegion.right);
		double viewY = changeCoordinates(y, (int) screenRegion.bottom, (int) screenRegion.top, viewRegion.bottom, viewRegion.top);
		viewY = viewRegion.bottom + viewRegion.top - viewY;
		return new Point(viewX, viewY);
	}
	
	/**
	 * Translate (move) a point by some number of pixels in the x and y directions.
	 * @param point - the point to move, in abstract coordinates
	 * @param x - shift to the right this many pixels
	 * @param y - shift down this many pixels
	 * @return - the translated point
	 */
	public Point translate(Point point, int x, int y) {
		return makePoint(getX(point) + x, getY(point) + y);
	}
	
	/**
	 * Find the distance between the two points, in screen coordinates.
	 */
	public int screenDistance(Point A, Point B) {
		int x1 = getX(A);
		int y1 = getY(A);
		int x2 = getX(B);
		int y2 = getY(B);
		return (int) Math.sqrt((x1 - x2)*(x1 - x2) + (y1 - y2)*(y1 - y2));
	}
	
	/**
	 * Find the distance between two points, in view coordinates.
	 */
	public double viewDistance(Point A, Point B) {
		double x1 = A.getX();
		double y1 = A.getY();
		double x2 = B.getX();
		double y2 = B.getY();
		return Math.sqrt((x1 - x2)*(x1 - x2) + (y1 - y2)*(y1 - y2));
	}
	
	// private
	
	private int changeCoordinates(double x, double fromMin, double fromMax, int toMin, int toMax) {
		if (fromMax - fromMin == 0)
			return (int) Math.round((toMin + toMax)/2);
		double f = (x - fromMin)/(fromMax - fromMin);
		return (int) Math.round(toMin + f*(toMax - toMin));
	}
	
	private double changeCoordinates(int x, int fromMin, int fromMax, double toMin, double toMax) {
		if (fromMax - fromMin == 0)
			return (toMin + toMax)/2.0;
		double f = (double) (x - fromMin)/(fromMax - fromMin);
		return toMin + f*(toMax - toMin);
	}
}
