package phyApps.common.graph;

//TODO:

//- Add labeled points to Graph
//- Make resizeable!

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Panel;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.swing.JPanel;

import phyApps.common.graph.Point;

/**
 * A Graph has several plots. It draws them and updates automatically when they change.
 * Each Plot takes up a certain rectangular region. The graph adjusts so that all Regions are shown.
 * @author JLP
 */
@SuppressWarnings("serial")
public class Graph extends JPanel {
	
	protected CoordinateSystem space;
	protected int border = 40;
	protected boolean fixedAspectRatio; // If true, the aspect ratio is fixed to be square (i.e. 1:1)
	protected List<Plot> plots = new CopyOnWriteArrayList<Plot>();
	protected static int POINT_RADIUS = 5;
	
	protected Color BACKGROUND = new Color(240, 240, 230);
	
	// Manual double buffering.
	//private Image buffer;
	//private Graphics bufferGraphics;
	
	// API:
	
	public Graph(Dimension size) {
		setBackground(BACKGROUND);
		setMinimumSize(new Dimension(0, 0));
		setPreferredSize(size);
		setSize(size);
		space = new CoordinateSystem(determineRegion(),
				new Region(border, border, size.getWidth() - border, size.getHeight() - border));
		this.setDoubleBuffered(true);
		fixedAspectRatio = false;
	}
	
	public Graph(Dimension size, Panel panel) {
		this(size);
		panel.add(this);
	}
	
	/**
	 * Draw all of the plots.
	 * Call this after changes have been made to the plots.
	 */
	public void draw() {
		repaint();
		//doubleBufferedDraw(getGraphics());
	}
	
	/**
	 * Add a plot to be drawn.
	 * @param plot
	 */
	public void add(Plot plot) {
		plots.add(plot);
		plot.setup(this);
		repaint();
	}
	
	/**
	 * Remove a plot from the graph.
	 * @param plot - the plot to remove
	 */
	public void remove(Plot plot) {
		plots.remove(plot);
		repaint();
	}
	
	/**
	 * Clear the graph of all plots.
	 */
	public void clear() {
		plots.clear();
		repaint();
	}
	
	/**
	 * Set the amount of whitespace around the graph.
	 * @param borderSize - size of the border, in pixels
	 */
	public void setBorderSize(int borderSize) {
		border = borderSize;
	}
	
	public CoordinateSystem getCoordinateSpace() {
		return space;
	}
	
	public Region getView() {
		return space.viewRegion;
	}
	
	public Region getScreen() {
		return space.screenRegion;
	}
	
	/**
	 * Fix the aspect ratio.
	 * @param b - if true, the aspect ratio is forced to be 1:1.
	 */
	public void fixAspectRatio(boolean b) {
		this.fixedAspectRatio = b;
	}
	
	
	// for Plots:
	
	/**
	 * Adjust the bounds to encompass the bounds (Regions) of all of the plots.
	 * A plot should call this when its boundary has changed.
	 * (Calling repaint() will suffice.)
	 */
	public void adjust() {
		adjust(getSize());
	}
	
	public void adjust(Dimension size) {
		if (!size.equals(getSize()))
			setSize(size);
		if (space != null && !space.viewRegion.equals(determineRegion()))
			space = new CoordinateSystem(determineRegion(),
					new Region(border, border, size.getWidth() - border, size.getHeight() - border));
	}
	
	protected Region determineRegion() {
		Region region = null;
		
		for (Plot plot : plots)
			if (!plot.isEmpty())
				region = Region.combine(region, plot.getRegion());
		
		// Make sure that *some* region is shown.
		if (region == null)
			region = new Region(0, 0, 1, 1);
		
		// Make sure the region isn't just a point.
		if (region.left == region.right)
			region.right += 1.0;
		if (region.bottom == region.top)
			region.top += 1.0;
		
		// If the aspect ratio is fixed, extend the region to the right or up as necessary.
		if (fixedAspectRatio) {
			double rw = Math.abs(region.right - region.left);
			double rh = Math.abs(region.top - region.bottom);
			double w = getSize().getWidth();
			double h = getSize().getHeight();
			
			if (rw != 0 && rh != 0 && w != 0 && h != 0) {
				if (h * rw > w * rh)
					region.top = region.bottom + rw*h/w;
				else
					region.right = region.left + rh*w/h;
			}
		}
		
		// Add a little bit to the region.
		double x = (region.right - region.left)*0.01;
		double y = (region.top - region.bottom)*0.01;
		region.left -= x;
		region.right += x;
		region.top += y;
		region.bottom -= y;
		
		return region;
	}
	
	// Drawing functions.
	
	public void drawLine(Graphics g, Point start, Point end, Color color) {
		drawLine(g, start, end, color, 1.0);
	}
	
	public void drawLine(Graphics g, Point start, Point end, Color color, double thickness) {
		Graphics2D g2 = (Graphics2D) g;
		g2.setStroke(new BasicStroke((float) thickness));
		g2.setColor(color);
		g2.drawLine(space.getX(start),  space.getY(start), space.getX(end),  space.getY(end));
	}
	
	public void fillBackground(Graphics g, Color color) {
		g.setColor(color);
		g.fillRect(0, 0, getWidth(), getHeight());
	}
	
	public void fillRect(Graphics g, Point lowLeft, Point upRight, Color color) {
		g.setColor(color);
		g.fillRect(space.getX(lowLeft), space.getY(upRight),
				space.getX(upRight) - space.getX(lowLeft),
				space.getY(lowLeft) - space.getY(upRight));
	}
	
	public void fillCircle(Graphics g, Point center, Point pointOnCircle, Color color) {
		int radius = space.screenDistance(center, pointOnCircle);
		g.setColor(color);
		g.fillOval(space.getX(center), space.getY(center), radius, radius);
	}
	
	public void fillPolygon(Graphics g, Point[] points, Color color) {
		int numPoints = points.length;
		int[] xPoints = new int[numPoints];
		int[] yPoints = new int[numPoints];
		
		for (int i=0; i<numPoints; i++) {
			xPoints[i] = space.getX(points[i]);
			yPoints[i] = space.getY(points[i]);
		}
		
		g.setColor(color);
		g.fillPolygon(xPoints, yPoints, numPoints);
	}
	
	public void drawPoint(Graphics g, Point point, Color color, int radius) {
		g.setColor(color);
		g.fillOval(space.getX(point) - radius, space.getY(point) - radius, 2*radius, 2*radius);
	}
	
	public void drawString(Graphics g, String text, Point location, Font font) {
		if (font != null)
			g.setFont(font);
		g.setColor(Color.BLACK);
		g.drawString(text, space.getX(location), space.getY(location));
	}
	
	public void drawImage(Graphics g, Image image, Point center) {
		int width = image.getWidth(null); // May be -1 if the image hasn't loaded. That's OK - just draw it at the wrong location.
		int height = image.getHeight(null);
		int x = space.getX(center);
		int y = space.getY(center);
		g.drawImage(image, x - width/2, y - height/2, null);
	}
	
	// Convenience functionality.
	// (There's no better place to put this function.)
	
	public static String formatDouble(double number) {
		return formatDouble(number, number/10.0);
	}
	
	/**
	 * Print a double to limited precision, while keeping the output string as short as possible.
	 * @param number - the number to format.
	 * @param scale - how precise you want the output to be.
	 * If scale = 0.03, the output will have two places after the decimal.
	 * @return - a *short* formatted representation of number.
	 */
	public static String formatDouble(double number, double scale) {
		return formatDouble(number, scale, null);
	}
	
	/**
	 * The same as formatDouble(double, double), but appends the units string to the end of the number.
	 */
	public static String formatDouble(double number, double scale, String units) {
		if (units == null)
			units = "";
		if (number == -0.0 || number == 0.0) {
			return Integer.toString(0) + units;
		}
		int places = Math.max(0, (int) -Math.floor(Math.log10(scale)));
		if (places > 10)
			places = 10;
		String formattedNumber = String.format("%." + Integer.toString(places) + "f", number);
		formattedNumber += units;
		return formattedNumber;
	}
	
	
	// Internal:
	/*
	private void doubleBufferedDraw(Graphics screen)
	{
		buffer = createImage(this.getSize().width, this.getSize().height);
		if (buffer == null)
			return;
		bufferGraphics = buffer.getGraphics();

		// Draw everything on the buffer.
		draw(bufferGraphics);
		
		// Move the buffer to the screen.
		screen.drawImage(buffer, 0, 0, this);
	}
	*/
	
	private void draw(Graphics graphics) {
		graphics.setColor(getBackground());
		graphics.fillRect(0, 0, getWidth(), getHeight());
		
		adjust();
		for (Plot plot : plots) {
			if (!plot.isEmpty())
				plot.draw(graphics);
		}
	}
	
	public void paint(Graphics g) {
		//doubleBufferedDraw(g);
		draw(g);
	}
	
	public void setBounds(int x, int y, int width, int height) {
		if (space != null)
			space.viewRegion = new Region(border, border, width - border, height - border);
		super.setBounds(x, y, width, height);
	}
}
