package phyApps.projectile;

import java.awt.Color;
import java.util.Timer;
import java.util.TimerTask;

import phyApps.common.graph.CurvePlot;
import phyApps.common.graph.Graph;
import phyApps.common.graph.LocationPlot;
import phyApps.common.graph.Point;

interface ProjectileObserver {
	public void update(Point newLocation); // Called every tick with the new location.
	public void done(); // Called if the projectile hits the ground.
}

/**
 * This class is responsible for drawing a projectile.
 * @author JLP
 */
public class Projectile {
	
	private Graph graph;
	private ProjectileObserver observer;
	
	private Timer timer;
	private TimerTask task;
	private LocationPlot locationDot; // A dot on the graph where the projectile currently is.
	private CurvePlot trail; // A curve on the graph showing where the projectile has been.
	private Trajectory trajectory; // The projectile's trajectory (where it is going).
	
	private static Color colors[] = {Color.blue, Color.red,
		Color.green, Color.orange, Color.cyan};
	private static int colorIndex;
	
	private double time; // The amount of time the projectile has been moving, in seconds.
	private static int period = 50; // The amount of time between ticks of the timer, in ms.
	
	public Projectile(Graph graph, Point location, LocationPlot plot, Point velocity) {
		this(graph, location, plot, velocity, null);
	}

	
	public Projectile(Graph graph, Point location, LocationPlot plot, Point velocity, ProjectileObserver observer) {
		this.graph = graph;
		this.observer = observer;
		this.trajectory = new Trajectory(location, velocity);
		
		Color color = nextColor();
		locationDot = plot;
		locationDot.move(location);
		trail = new CurvePlot(graph, color);
		trail.setRegion(trajectory.getRegion());
	}
	
	// Cycle through the available colors.
	private static Color nextColor() {
		Color color = colors[colorIndex];
		colorIndex++;
		if (colorIndex >= colors.length)
			colorIndex = 0;
		return color;
	}
	
	public void start() {
		time = 0.0;
		timer = new Timer();
		task = new ProjectileTimerTask();
		timer.scheduleAtFixedRate(task, 0, period);
	}
	
	public void stop() {
		if (timer != null)
			timer.cancel();
		graph.remove(locationDot);
	}
	
	public boolean isMoving() {
		return !trajectory.isDone(time);
	}
	
	public void remove() {
		stop();
		graph.remove(locationDot);
		graph.remove(trail);
	}
	
	public double getMaxHeight() {
		return trajectory.getPeak();
	}
	
	public double getDistance() {
		return trajectory.getDisplacement();
	}
	
	public void tick() {
		time += period/1000.0;
		show(time);
		if (trajectory.isDone(time)) {
			if (observer != null)
				observer.done();
			stop();
		}
	}
	
	private void show(double time) {
		Point location = trajectory.getLocation(time);
		if (observer != null)
			observer.update(location);
		locationDot.move(location);
		trail.setPoints(trajectory.getTrail(time));
		graph.draw();
	}
	
	public class ProjectileTimerTask extends TimerTask {
		public void run() {
			tick();
		}
	}
}
