package phyApps.projectile;

import java.awt.Button;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Label;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JOptionPane;

import phyApps.common.ControlPanel;
import phyApps.common.ControlPanelListener;
import phyApps.common.Distribution;
import phyApps.common.graph.ArrowPlot;
import phyApps.common.graph.AxisPlot;
import phyApps.common.graph.Graph;
import phyApps.common.graph.LocationPlot;
import phyApps.common.graph.Plot;
import phyApps.common.graph.Point;
import phyApps.common.graph.PointPlot;
import phyApps.common.graph.Region;
import phyApps.common.graph.TargetPlot;
import phyApps.common.mvc.PhyApplet;

@SuppressWarnings("serial")
public class ProjectileGame extends PhyApplet {
	
	@Override
	public void init() {
		ProjectileGameView view = new ProjectileGameView();
		ProjectileGameController controller = new ProjectileGameController(view.getGraph());
		
		PhyApplet.layoutApplet(this, controller.getComponent(), view.getGraph());
	}
	
	public static void main(String[] args) {
		PhyApplet.runApplet(new ProjectileGame(), "Dart Toss");
	}
}


class ProjectileGameView {

	private Graph graph;
	
	public ProjectileGameView() {
		graph = new Graph(new Dimension(500, 500));
		graph.fixAspectRatio(true);
		AxisPlot plot = new AxisPlot(graph);
		plot.setXUnit("m");
		plot.setYUnit("m");
		plot.toggleScale(true);
		plot.toggleNumbers(true);
		plot.toggleGrid(true);
	}
	
	public Graph getGraph() {
		return graph;
	}
}


class ProjectileGameController implements ActionListener, ProjectileObserver, ControlPanelListener {
	
	private ControlPanel controlPanel;
	private Graph graph;
	private Projectile projectile;
	private List<Projectile> history;
	
	private Button launchButton;
	private Button nextButton;
	
	private LocationPlot target;
	private ArrowPlot launchArrow;
	
	private GoatPlot goatPlot;
	private int level; // The current level number.
	private int goats; // The number of goats left.
	
	private int STARTING_GOATS = 3;
	private double targetSize; // How close the goat has to get to the box.
	
	private static double ARROW_SIZE = 2.0;
	
	public ProjectileGameController(Graph graph) {
		this.controlPanel = new ControlPanel();
		this.graph = graph;
		this.history = new ArrayList<Projectile>();
		this.goatPlot = new GoatPlot(graph);
		this.launchArrow = new ArrowPlot(graph, Color.BLUE);
		restart();
	}
	
	public Component getComponent() {
		return controlPanel;
	}
	
	public void actionPerformed(ActionEvent event) {
		drawLaunchArrow();
		String command = event.getActionCommand();
		if (command.equals("launch"))
			launch();
		else if (command.equals("clear"))
			clear();
		else if (command.equals("restart"))
			restart();
		else if (command.equals("next"))
			next();
	}
	
	public void update(Point newLocation) {
		if (near(newLocation, target.getLocation()))
			win();
	}
	
	public void done() {
		miss();
	}
	
	private void loadLevel() {
		switch (level) {
		case 1:
			loadLevel(Level.level1);
			break;
		case 2:
			loadLevel(Level.level2);
			break;
		case 3:
			loadLevel(Level.level3);
			break;
		case 4:
			loadLevel(Level.level4);
			break;
		case 5:
			loadLevel(Level.level5);
			break;
		case 6:
			loadLevel(Level.level6);
			break;
		default:
			winGame();
		}
	}
	
	private void loadLevel(Level level) {
		// Clear the old level.
		for (Projectile p : history)
			p.remove();
		history.clear();
		
		// Construct the panel.
		controlPanel.clear();
		
		controlPanel.addText("Level " + Integer.toString(this.level),
				Label.LEFT,
				new Font("sanserif", Font.BOLD, 14));
		controlPanel.addLabel("goats", "Darts left");
		controlPanel.setValue("goats", goats);
		
		level.addControls(controlPanel);
		controlPanel.addListener("xpos", this);
		controlPanel.addListener("ypos", this);
		controlPanel.addListener("xvel", this);
		controlPanel.addListener("yvel", this);
		
		launchButton = controlPanel.addButton("launch", "Launch", this);
		controlPanel.addButton("clear", "Clear", this);
		controlPanel.addButton("restart", "Restart", this);
		nextButton = controlPanel.addButton("next", "Next Level", this);
		nextButton.setEnabled(false);
		controlPanel.validate();
		
		// Place the target.
		if (target != null)
			graph.remove(target);
		Point targetLocation = level.getTargetLocation();
		targetSize = level.getTargetSize();
		target = new TargetPlot(graph, targetSize, Color.RED, Color.WHITE);
		double x = targetLocation.getX();
		double y = targetLocation.getY();
		target.move(targetLocation);
		target.setRegion(new Region(0, 0, 1.5*x, 1.5*y));
		
		drawLaunchArrow();
	}
	
	private void launch() {
		projectile = new Projectile(graph,
				new Point(getXPos(), getYPos()),
				new PointPlot(graph, Color.BLUE),
				new Point(getXVel(), getYVel()),
				this);
		history.add(projectile);
		projectile.start();
	}
	
	private void clear() {
		for (Projectile p : history) {
			if (p.isMoving())
				miss();
			p.remove();
		}
		history.clear();
	}
	
	private double getXPos() {
		return controlPanel.getValue("xpos");
	}
	
	private double getYPos() {
		return controlPanel.getValue("ypos");
	}
	
	private double getXVel() {
		return controlPanel.getValue("xvel");
	}
	
	private double getYVel() {
		return controlPanel.getValue("yvel");
	}
	
	private boolean near(Point p1, Point p2) {
		return graph.getCoordinateSpace().viewDistance(p1, p2) < targetSize;
	}
	
	private void miss() {
		setGoats(goats - 1);
	}
	
	private void setGoats(int numGoats) {
		goats = numGoats;
		if (goats <= 0) {
			goats = 0;
			loseGame();
		}
		else {
			controlPanel.setValue("goats", goats);
		}
		goatPlot.setGoats(goats);
		graph.draw();
	}
	
	// Win a level.
	private void win() {
		projectile.stop();
		graph.draw();
		
		launchButton.setEnabled(false);
		nextButton.setEnabled(true);
	}
	
	// Win the game.
	private void winGame() {
		String message;
		if (goats < STARTING_GOATS)
			message = String.format("You won after losing %d darts.", STARTING_GOATS - goats);
		else
			message = "Perfect game!";
		
		JOptionPane.showMessageDialog(null, message);
		restart();
	}
	
	// Lose the game.
	private void loseGame() {
		JOptionPane.showMessageDialog(null, "You ran out of darts.");
		restart();
	}
	
	// Restart from the first level.
	private void restart() {
		level = 1;
		loadLevel();
		setGoats(STARTING_GOATS);
	}
	
	// Continue to the next level.
	private void next() {
		level++;
		loadLevel();
	}
	
	public void drawLaunchArrow() {
		double x = controlPanel.getValue("xpos");
		double y = controlPanel.getValue("ypos");
		double xvel = controlPanel.getValue("xvel");
		double yvel = controlPanel.getValue("yvel");
		Point start = new Point(x, y);
		Point end = graph.getCoordinateSpace().translate(start, (int) (xvel*ARROW_SIZE), (int) (-yvel*ARROW_SIZE));
		launchArrow.move(start, end);
		graph.draw();
	}

	public void valueChanged(String field, double value) {
		drawLaunchArrow();
	}
}

class Level {
	
	private static boolean[] level1mutability = {false, true, false, false};
	private static boolean[] level2mutability = {false, true, false, true};
	private static boolean[] level3mutability = {false, false, true, true};
	private static boolean[] level4mutability = {false, true, false, true};
	private static boolean[] level5mutability = {false, true, false, false};
	private static boolean[] level6mutability = {false, false, true, true};
	
	public static Level level1 = new Level(level1mutability, 3.0);
	public static Level level2 = new Level(level2mutability, 2.5);
	public static Level level3 = new Level(level3mutability, 2.0);
	public static Level level4 = new Level(level4mutability, 1.5);
	public static Level level5 = new Level(level5mutability, 1.0);
	public static Level level6 = new Level(level6mutability, 1.0);
	
	private String[][] ATTRIBUTES = {
			{"xpos", "X Position"},
			{"ypos", "Y Position"},
			{"xvel", "X Velocity"},
			{"yvel", "Y Velocity"}};
	private Distribution[] RANGES = { // of the form {min, max, inc}
			new Distribution(0, 0),
			new Distribution(10, 20, 5),
			new Distribution(10, 20, 5),
			new Distribution(10, 20, 5)};
	private Distribution[] TARGET_RANGE = { // of the form {min, max, inc}
			new Distribution(10, 30, 5),
			new Distribution(10, 30, 5)};
	
	private boolean[] mutability;
	private double targetSize;
	
	public Level(boolean[] attributeMutabilityList, double targetSize) {
		this.mutability = attributeMutabilityList;
		this.targetSize = targetSize;
	}
	
	public void addControls(ControlPanel panel) {
		for (int i = 0; i < ATTRIBUTES.length; i++) {
			if (mutability[i]) {
				panel.addTextField(ATTRIBUTES[i][0], ATTRIBUTES[i][1], 0.0);
			}
			else {
				panel.addLabel(ATTRIBUTES[i][0], ATTRIBUTES[i][1]);
				panel.setValue(ATTRIBUTES[i][0], RANGES[i].pick());
			}
		}
	}
	
	public Point getTargetLocation() {
		return new Point(TARGET_RANGE[0].pick(), TARGET_RANGE[1].pick());
	}
	
	public double getTargetSize() {
		return targetSize;
	}
}


class GoatPlot extends Plot {

	private int goats = 0;

	private int radius = 10;
	private int border = 5;
	private int x = radius + border;
	private int y = radius + border;
	private int xInc = 2*radius + border;
	
	public GoatPlot(Graph graph) {
		super(graph);
	}
	
	public void setGoats(int numGoats) {
		goats = numGoats;
	}

	public Region determineRegion() {
		return null;
	}

	public void draw(Graphics g) {
		for (int i=0; i<goats; i++) {
			Point location = graph.getCoordinateSpace().makePoint(x + i*xInc, y);
			graph.drawPoint(g, location, Color.BLUE, radius);
		}
	}

	public boolean isEmpty() {
		return false;
	}
}
